from __future__ import annotations

# These annotations are based off the standards:
# https://github.com/opencontainers/image-spec/blob/main/annotations.md
# https://github.com/CloudNativeAI/model-spec

# ArtifactTypeModelManifest specifies the media type for a model manifest.
ArtifactTypeModelManifest = "application/vnd.cncf.model.manifest.v1+json"

# ArtifactTypeModelLayer is the media type used for layers referenced by the
# manifest.
ArtifactTypeModelLayer = "application/vnd.cncf.model.layer.v1.tar"

# ArtifactTypeModelLayerGzip is the media type used for gzipped layers
# referenced by the manifest.
ArtifactTypeModelLayerGzip = "application/vnd.cncf.model.layer.v1.tar+gzip"

# AnnotationCreated is the annotation key for the date and time on which the
# model was built (date-time string as defined by RFC 3339).
AnnotationCreated = "org.opencontainers.image.created"

# AnnotationAuthors is the annotation key for the contact details of the people
# or organization responsible for the model (freeform string).
AnnotationAuthors = "org.opencontainers.image.authors"

# AnnotationURL is the annotation key for the URL to find more information on
# the artifact.
AnnotationURL = "org.opencontainers.image.url"

# AnnotationDocumentation is the annotation key for the URL to get documentation
# on the artifact.
AnnotationDocumentation = "org.opencontainers.image.documentation"

# AnnotationSource is the annotation key for the URL to get source code for
# building the artifact.
AnnotationSource = "org.opencontainers.image.source"

# AnnotationVersion is the annotation key for the version of the packaged
# software.
# The version MAY match a label or tag in the source code repository.
# The version MAY be Semantic versioning-compatible.
AnnotationVersion = "org.opencontainers.image.version"

# AnnotationRevision is the annotation key for the source control revision
# identifier for the packaged artifact.
AnnotationRevision = "org.opencontainers.image.revision"

# AnnotationVendor is the annotation key for the name of the distributing
# entity, organization or individual.
AnnotationVendor = "org.opencontainers.image.vendor"

# AnnotationLicenses is the annotation key for the license(s) under which
# contained software is distributed as an SPDX License Expression.
AnnotationLicenses = "org.opencontainers.image.licenses"

# AnnotationRefName is the annotation key for the name of the reference for a
# target.
# SHOULD only be considered valid when on descriptors on `index.json` within
# artifact layout.
AnnotationRefName = "org.opencontainers.image.ref.name"

# AnnotationTitle is the annotation key for the human-readable title of the
# artifact.
AnnotationTitle = "org.opencontainers.image.title"

# AnnotationDescription is the annotation key for the human-readable description
# of the software packaged in the artifact.
AnnotationDescription = "org.opencontainers.image.description"

# AnnotationBaseImageDigest is the annotation key for the digest of the image's
# base image.
AnnotationBaseImageDigest = "org.opencontainers.image.base.digest"

# AnnotationBaseImageName is the annotation key for the image reference of the
# image's base image.
AnnotationBaseImageName = "org.opencontainers.image.base.name"

# DEPRECATED: Migrate to AnnotationFilepath
# AnnotationModel is the annotation key for the layer is a model file (boolean),
# such as `true` or `false`.
AnnotationModel = "org.cncf.model.model"

# AnnotationFilepath is the annotation key for the file path of the layer.
AnnotationFilepath = "org.cncf.model.filepath"



from __future__ import annotations

import configparser
import os
import re
import sys
import sysconfig
from typing import Optional


class Shortnames:
    """Shortnames utility class"""

    shortnames: dict[str, str] = {}
    config_sources: dict[str, str] = {}

    def __init__(self):
        self.shortnames = {}
        self.config_sources = {}
        data_path = sysconfig.get_path("data")
        file_paths = [
            "./shortnames/shortnames.conf",  # for development
            "./shortnames.conf",  # for development
            f"{data_path}/share/ramalama/shortnames.conf",
        ]

        if os.name == 'nt':
            # Windows-specific paths using APPDATA and LOCALAPPDATA
            appdata = os.getenv("APPDATA", os.path.expanduser("~/AppData/Roaming"))
            localappdata = os.getenv("LOCALAPPDATA", os.path.expanduser("~/AppData/Local"))
            file_paths.extend(
                [
                    os.path.join(localappdata, "ramalama", "shortnames.conf"),
                    os.path.join(appdata, "ramalama", "shortnames.conf"),
                ]
            )
        else:
            # Unix-specific paths using XDG conventions
            file_paths.extend(
                [
                    os.path.expanduser(
                        os.path.join(os.getenv("XDG_CONFIG_HOME", "~/.config"), "ramalama/shortnames.conf")
                    ),
                    os.path.expanduser(
                        os.path.join(os.getenv("XDG_DATA_HOME", "~/.local/share"), "ramalama/shortnames.conf")
                    ),
                    os.path.expanduser("~/.local/pipx/venvs/ramalama/share/ramalama/shortnames.conf"),
                    "/etc/ramalama/shortnames.conf",
                    "/usr/share/ramalama/shortnames.conf",
                    "/usr/local/share/ramalama/shortnames.conf",
                ]
            )

        self.paths = []
        for file_path in file_paths:
            config = configparser.ConfigParser(delimiters="=")
            config.read(file_path)
            if "shortnames" in config:
                real_path = os.path.realpath(file_path)
                self.paths.append(real_path)
                for key, value in config["shortnames"].items():
                    name = self._strip_quotes(key)
                    target = self._strip_quotes(value)
                    self.shortnames[name] = target
                    self.config_sources[name] = real_path

        self._targets: dict[str, list[str]] = {}
        for name, target in self.shortnames.items():
            self._targets.setdefault(target, []).append(name)

    def _strip_quotes(self, s) -> str:
        return s.strip("'\"")

    def resolve(self, model: str) -> str:
        return self.shortnames.get(model, model)

    @staticmethod
    def sort_file(file_path: str, *, check: bool = False) -> bool:
        """Sort shortnames.conf entries by key, preserving header comments.

        Returns True if file is sorted. With check=True, returns False instead of writing.
        """

        with open(file_path) as f:
            lines = f.readlines()

        header: list[str] = []
        entries: list[str] = []
        in_entries = False
        for line in lines:
            if not in_entries:
                header.append(line)
                if line.strip().lower() == '[shortnames]':
                    in_entries = True
            else:
                entries.append(line)

        def is_mapping(line: str) -> bool:
            return line.strip().startswith('"')

        def parse_key(line: str) -> str:
            s = line.strip()
            return s[1 : s.index('"', 1)]

        def parse_tag_as_numeric(tag: str) -> float | None:
            multipliers = {'b': 1e9, 'm': 1e6, 'k': 1e3}
            m = re.search(r'(\d+)x(\d+\.?\d*)([bmk])\b', tag, re.IGNORECASE)
            if m:
                return float(m.group(1)) * float(m.group(2)) * multipliers[m.group(3).lower()]
            m = re.search(r'(\d+\.?\d*)([bmk])\b', tag, re.IGNORECASE)
            if m:
                return float(m.group(1)) * multipliers[m.group(2).lower()]
            m = re.search(r'(\d+\.?\d*)', tag)
            if m:
                return float(m.group(1))
            return None

        def entry_sort_key(line: str) -> tuple:
            key = parse_key(line)
            tag = key.split(':', 1)[1] if ':' in key else ''
            num = parse_tag_as_numeric(tag)
            return (num is not None, num, tag if num is None else '')

        def entry_name(line: str) -> str:
            key = parse_key(line)
            return key.split(':')[0] if ':' in key else key

        mappings = [line for line in entries if is_mapping(line)]
        # Pass 1: sort by tag — non-numeric tags first (lexically), then numeric tags by value
        sorted_mappings = sorted(mappings, key=entry_sort_key)
        # Pass 2: stable sort lexically by name
        sorted_mappings = sorted(sorted_mappings, key=entry_name)

        if mappings == sorted_mappings:
            return True

        if check:
            return False

        it = iter(sorted_mappings)
        result = [next(it) if is_mapping(line) else line for line in entries]

        with open(file_path, 'w') as f:
            f.writelines(header)
            f.writelines(result)

        return True

    def lookup(self, model: str) -> Optional[str]:
        """Return the preferred shortname for a resolved model URI, if any."""
        names = self._targets.get(model)
        if not names:
            return None
        return min(names)


if __name__ == "__main__":
    args = [a for a in sys.argv[1:] if not a.startswith("-")]
    check = "--check" in sys.argv[1:]
    if len(args) != 1:
        print(f"Usage: {sys.argv[0]} [--check] <conf-file>", file=sys.stderr)
        sys.exit(2)
    if not Shortnames.sort_file(args[0], check=check):
        print(f"ERROR: {args[0]} is not sorted by key. Run 'make format' to fix.", file=sys.stderr)
        sys.exit(1)



from __future__ import annotations

import os
import platform

from ramalama.config import ActiveConfig
from ramalama.console import EMOJI


def default_prefix() -> str:
    if not EMOJI:
        return "> "

    config = ActiveConfig()
    if config.prefix:
        return config.prefix

    if platform.system() == "Darwin":
        return "🍏 > "

    engine = config.engine
    if engine:
        engine_name = os.path.basename(engine)
        if engine_name == "podman":
            return "\U0001f9ad > "
        if engine_name == "docker":
            return "\U0001f40b > "

    return "\U0001f999 > "



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import asdict, fields
from http.client import HTTPConnection
from typing import Any, Optional, Type


class RuntimePlugin(ABC):
    config_type: Optional[Type] = None

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    def config_key(self) -> str:
        return self.name.replace(".", "_").replace("-", "_")

    def get_runtime_config(self, config: Any) -> Any:
        """Return the typed runtime config, hydrated from config.runtimes."""
        if self.config_type is None:
            return None
        raw = config.runtimes.get(self.config_key, {})
        if not isinstance(raw, Mapping):
            raise ValueError(f"config.runtimes.{self.config_key} must be a mapping, got {type(raw).__name__}")
        known = {f.name for f in fields(self.config_type)}
        return self.config_type(**{k: v for k, v in raw.items() if k in known})

    def sync_args_to_runtime_config(self, args: argparse.Namespace, config: Any) -> None:
        """Sync CLI args into the runtime config section."""
        if self.config_type is None:
            return
        rt_config = self.get_runtime_config(config)
        rt_fields = {f.name for f in fields(rt_config)}
        for arg_name in vars(args).keys() & rt_fields:
            if getattr(args, arg_name) != getattr(rt_config, arg_name):
                setattr(rt_config, arg_name, getattr(args, arg_name))
        config.runtimes[self.config_key] = asdict(rt_config)

    def get_container_image(self, config: Any, gpu_type: str) -> Optional[str]:
        return None

    def register_subcommands(self, subparsers: "argparse._SubParsersAction") -> None:
        """Register runtime-specific subcommand parsers.

        Override to add subcommands that only apply to this runtime.
        Called from configure_subcommands() after universal commands are registered.
        """
        pass

    def post_process_args(self, args: argparse.Namespace) -> None:
        """Mutate and validate args after parsing. Override in concrete plugins."""
        pass

    def handle_subcommand(self, command: str, args: argparse.Namespace) -> list[str]:
        """Handle the given subcommand. Override in concrete plugins."""
        raise NotImplementedError(f"{self.name} plugin does not implement handle_subcommand()")

    @property
    def service_ready_check_timeout(self) -> int:
        """Seconds to wait for the runtime server to become ready."""
        return 180

    def service_ready_check(self, conn: HTTPConnection, args: Any, model_name: Optional[str] = None) -> bool:
        """Check if the service is ready to receive requests."""
        return True


class InferenceRuntimePlugin(RuntimePlugin, ABC):
    """Abstract base class for inference runtime plugins."""

    @property
    def chat_include_model_name(self) -> bool:
        """Whether to include the model name in chat API requests."""
        return True



#!/usr/bin/env python3
from __future__ import annotations

import _thread
import cmd
import copy
import itertools
import json
import os
import signal
import subprocess
import sys
import threading
import time
import urllib.error
import urllib.request
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from ramalama.arg_types import ChatArgsType
from ramalama.chat_providers import ChatProvider, ChatRequestOptions
from ramalama.chat_providers.openai import OpenAICompletionsChatProvider
from ramalama.chat_utils import (
    AssistantMessage,
    ChatMessageType,
    SystemMessage,
    ToolMessage,
    UserMessage,
    sanitize_for_terminal,
    stream_response,
)
from ramalama.common import perror
from ramalama.config import ActiveConfig
from ramalama.console import should_colorize
from ramalama.engine import stop_container
from ramalama.file_loaders.file_manager import OpanAIChatAPIMessageBuilder
from ramalama.logger import logger
from ramalama.mcp.mcp_agent import LLMAgent
from ramalama.mcp.mcp_client import PureMCPClient
from ramalama.plugins.interface import InferenceRuntimePlugin
from ramalama.plugins.loader import get_runtime
from ramalama.proxy_support import setup_proxy_support

# Setup proxy support on module import
setup_proxy_support()


def res(response, color):
    color_default = ""
    color_yellow = ""
    if (color == "auto" and should_colorize()) or color == "always":
        color_default = "\033[0m"
        color_yellow = "\033[33m"

    print("\r", end="")
    assistant_response = ""
    for line in response:
        line = line.decode("utf-8").strip()
        if line.startswith("data: {"):
            choice = ""

            json_line = json.loads(line[len("data: ") :])
            if "choices" in json_line and json_line["choices"]:
                choice = json_line["choices"][0]["delta"]
            if isinstance(choice, dict) and "content" in choice:
                content = choice["content"]
            else:
                continue

            if content:
                safe_content = sanitize_for_terminal(content)
                print(f"{color_yellow}{safe_content}{color_default}", end="", flush=True)
                assistant_response += content

    print("")
    return assistant_response


def add_api_key(args, headers=None):
    # static analyzers suggest for dict, this is a safer way of setting
    # a default value, rather than using the parameter directly
    headers = headers or {}
    if getattr(args, "api_key", None):
        api_key_min = 20
        if len(args.api_key) < api_key_min:
            perror("Warning: Provided API key is invalid.")

        headers["Authorization"] = f"Bearer {args.api_key}"

    return headers


@dataclass
class ChatOperationalArgs:
    name: Optional[str] = None
    keepalive: Optional[int] = None
    monitor: "Optional[ServerMonitor]" = None


class Spinner:
    def __init__(self, wait_time: float = 0.1):
        self._stop_event: threading.Event = threading.Event()
        self._thread: Optional[threading.Thread] = None
        self.wait_time = wait_time

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False

    def start(self) -> "Spinner":
        if not sys.stdout.isatty():
            return self

        if self._thread is not None:
            self.stop()

        self._thread = threading.Thread(target=self._spinner_loop, daemon=True)
        self._thread.start()
        return self

    def stop(self):
        if self._thread is None:
            return

        self._stop_event.set()
        self._thread.join(timeout=0.2)
        perror("\r", end="", flush=True)
        self._thread = None
        self._stop_event = threading.Event()

    def _spinner_loop(self):
        frames = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']

        for frame in itertools.cycle(frames):
            if self._stop_event.is_set():
                break
            perror(f"\r{frame}", end="", flush=True)
            self._stop_event.wait(self.wait_time)


class RamaLamaShell(cmd.Cmd):
    def __init__(
        self,
        args: ChatArgsType,
        operational_args: Optional[ChatOperationalArgs] = None,
        provider: Optional[ChatProvider] = None,
    ):
        if operational_args is None:
            operational_args = ChatOperationalArgs()

        super().__init__()
        self.conversation_history: list[ChatMessageType] = []
        self.args = args
        self.operational_args = operational_args
        self.request_in_process = False
        self.prompt = args.prefix
        self.provider = provider or OpenAICompletionsChatProvider(args.url, getattr(args, "api_key", None))
        self.url = self.provider.build_url()

        self.prep_rag_message()
        self.mcp_agent: Optional[LLMAgent] = None
        self.initialize_mcp()

        self.content: list[str] = []
        self.message_count = 0  # Track messages for summarization

    def do_help(self, args):
        """Display help information about available commands."""
        print("\nAvailable commands:")
        print("  /help, help, ?    - Show this help message")
        print("  /clear            - Clear conversation history")
        print("  /bye, exit        - Exit the chat session")
        if self.mcp_agent:
            print("  /tool [question]  - Manually select which MCP tool to use")
        print("  \\                 - End a line with backslash to continue on next line")
        print("  Ctrl+D            - Exit the chat session (EOF)")
        print()

    def prep_rag_message(self):
        if (context := getattr(self.args, 'rag', None)) is None:
            return

        builder = OpanAIChatAPIMessageBuilder()
        messages = builder.load(context)
        self.conversation_history.extend(messages)

    def _summarize_conversation(self):
        """Summarize the conversation history to prevent context growth."""
        if len(self.conversation_history) < 10:
            # Need at least a few messages to summarize
            return

        # Keep the first message (system/RAG context) and last 2 messages
        # Summarize everything in between
        first_msg = self.conversation_history[0]
        messages_to_summarize = self.conversation_history[1:-2]
        recent_msgs = self.conversation_history[-2:]

        if not messages_to_summarize:
            return

        # Create a summarization prompt
        conversation_text = "\n".join([self._format_message_for_summary(msg) for msg in messages_to_summarize])
        summary_prompt = UserMessage(
            text=(
                "Please provide a concise summary of the following conversation, "
                f"preserving key information and context:\n\n{conversation_text}\n\n"
                "Provide only the summary, without any preamble."
            )
        )

        # Make API call to get summary
        # Provide user feedback during summarization
        print("\nSummarizing conversation to reduce context size...", flush=True)
        try:
            req = self._make_api_request([summary_prompt], stream=False)

            with urllib.request.urlopen(req) as response:
                result = json.loads(response.read())
                summary = result['choices'][0]['message']['content']

                # Rebuild conversation history with summary
                new_history: list[ChatMessageType] = []
                if first_msg:
                    new_history.append(first_msg)

                # Add summary as a system message
                new_history.append(SystemMessage(text=f"Previous conversation summary: {summary}"))

                # Add recent messages
                new_history.extend(recent_msgs)

                self.conversation_history = new_history
                logger.debug(f"Summarized conversation: {len(messages_to_summarize)} messages -> 1 summary")

        except (urllib.error.URLError, json.JSONDecodeError, KeyError, IndexError) as e:
            logger.warning(f"Failed to summarize conversation: {e}")
            # On failure, just keep the conversation as-is
        finally:
            # Clear the "Summarizing..." message
            print("\r" + " " * 60 + "\r", end="", flush=True)

    def _check_and_summarize(self):
        """Check if conversation needs summarization and trigger it."""
        summarize_after = getattr(self.args, "summarize_after", 0)
        if summarize_after > 0:
            self.message_count += 2  # user + assistant messages
            if self.message_count >= summarize_after:
                self._summarize_conversation()
                self.message_count = 0  # Reset counter after summarization

    def _history_snapshot(self) -> list[dict[str, str]]:
        return [
            {"role": msg.role, "content": self._format_message_for_summary(msg)} for msg in self.conversation_history
        ]

    def _format_message_for_summary(self, msg: ChatMessageType) -> str:
        content = msg.text or ""
        if isinstance(msg, AssistantMessage):
            if msg.tool_calls:
                content += f"\n[tool_calls: {', '.join(call.name for call in msg.tool_calls)}]"

        if isinstance(msg, ToolMessage):
            content = f"\n[tool_response: {msg.text}]"

        return f"{msg.role}: {content}".strip()

    def _make_api_request(self, messages: Sequence[ChatMessageType], stream: bool = True):
        """Create a provider request for arbitrary message lists."""
        max_tokens = self.args.max_tokens if stream and getattr(self.args, "max_tokens", None) else None
        options = self._build_request_options(stream=stream, max_tokens=max_tokens)
        return self.provider.create_request(messages, options)

    def _resolve_model_name(self) -> Optional[str]:
        runtime = getattr(self.args, "runtime", None)
        if runtime:
            plugin = get_runtime(runtime)
            if isinstance(plugin, InferenceRuntimePlugin) and not plugin.chat_include_model_name:
                return None
        return getattr(self.args, "model", None)

    def _build_request_options(self, *, stream: bool, max_tokens: Optional[int]) -> ChatRequestOptions:
        temperature = getattr(self.args, "temp", None)
        if max_tokens is not None and max_tokens <= 0:
            max_tokens = None
        return ChatRequestOptions(
            model=self._resolve_model_name(),
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
        )

    def initialize_mcp(self):
        """Initialize MCP servers if specified."""
        if not hasattr(self.args, 'mcp') or not self.args.mcp:
            return

        try:
            # Create clients for each MCP server URL
            clients = []
            for url in self.args.mcp:
                try:
                    clients.append(PureMCPClient(url))
                    logger.debug(f"Created MCP client for {url}")
                except Exception as e:
                    perror(f"Failed to create MCP client for {url}: {e}")

            if clients:
                # Use the same LLM URL as the chat for the agent
                llm_url = self.args.url
                self.mcp_agent = LLMAgent(clients, llm_url, self.args.model, self.args)

                # Set up proper streaming callback for MCP agent
                def mcp_stream_callback(text):
                    color_default = ""
                    color_yellow = ""
                    if (self.args.color == "auto" and should_colorize()) or self.args.color == "always":
                        color_default = "\033[0m"
                        color_yellow = "\033[33m"
                    safe_text = sanitize_for_terminal(text)
                    print(f"{color_yellow}{safe_text}{color_default}", end="", flush=True)

                setattr(self.mcp_agent, "_stream_callback", mcp_stream_callback)

                # Initialize the agent and get available tools
                init_results, tools = self.mcp_agent.initialize()

                # Show connection summary (simplified)
                for i, result in enumerate(init_results):
                    server_name = result['result']['serverInfo']['name']
                    server_tools = [tool for tool in tools if tool.get('server') == server_name]
                    logger.debug(f"Connected to: {server_name}")
                    print(f"Found {len(server_tools)} tool(s) from {server_name}")

                self.do_help("")

        except Exception as e:
            perror(f"Failed to initialize MCP: {e}")
            logger.debug(f"MCP initialization error: {e}", exc_info=True)

    def _should_use_mcp(self, content: str) -> bool:
        """Determine if the request should be handled by MCP tools."""
        if not self.mcp_agent:
            return False
        return self.mcp_agent.should_use_tools(content, self._history_snapshot())

    def _handle_mcp_request(self, content: str) -> str:
        """Handle a request using MCP tools (multi-tool capable, automatic)."""
        try:
            assert self.mcp_agent
            # Automatic tool selection and argument generation
            results = self.mcp_agent.execute_task(content, manual=False, stream=True)

            # When streaming, results will be None since output is streamed directly
            if results is None:
                return ""  # Return empty string since content was already streamed
            # If multiple tools ran, join results into a single answer
            elif isinstance(results, list):
                combined = "\n\n".join(f"🔧 {r['tool']}: {r['output']}" for r in results)
                return f"Multi-tool execution:\n{combined}"
            else:
                return results
        except Exception as e:
            logger.debug(f"MCP request handling error: {e}", exc_info=True)
            return f"Error using MCP tools: {e}"

    def _handle_manual_tool_selection(self, content: str):
        if not self.mcp_agent or not self.mcp_agent.available_tools:
            perror("No MCP tools available.")
            return

        parts = content.strip().split(None, 1)
        question = parts[1] if len(parts) > 1 else ""

        selected_tools = self._select_tools()
        if not selected_tools:
            return

        responses = []
        for tool in selected_tools:
            response = self.mcp_agent.execute_specific_tool(question, tool['name'], manual=False)
            responses.append({"tool": tool['name'], "output": response})

        # Display results
        for r in responses:
            print(f"\n {r['tool']} -> {r['output']}")

        # Save to history
        self.conversation_history.append(UserMessage(text=f"/tool {question}"))
        self.conversation_history.append(AssistantMessage(text=str(responses)))

    def _select_tools(self):
        """Interactive multi-tool selection without prompting for arguments."""
        if not self.mcp_agent or not self.mcp_agent.available_tools:
            return None

        self.mcp_agent.print_tools()

        try:
            choice = input("\nSelect tool(s) (e.g. 1,2,3) or 'q' to cancel: ").strip()
            if choice.lower() == 'q':
                return None

            indices = [int(c.strip()) - 1 for c in choice.split(",")]
            selected = [
                self.mcp_agent.available_tools[i] for i in indices if 0 <= i < len(self.mcp_agent.available_tools)
            ]

            return selected

        except (ValueError, KeyboardInterrupt):
            perror("\nCancelled.")
            return None

    def handle_args(self, monitor):
        """Process command-line arguments and/or stdin input.

        Returns True if the program should exit, False if it should continue to interactive mode.
        """
        prompt = " ".join(self.args.ARGS) if self.args.ARGS else None
        stdin_was_pipe = False

        if not sys.stdin.isatty():
            stdin = sys.stdin.read()
            stdin_was_pipe = True
            if prompt:
                prompt += f"\n\n{stdin}"
            else:
                prompt = stdin

        if prompt:
            self.default(prompt)
            # Continue to interactive mode if --interactive flag is set
            if getattr(self.args, 'interactive', False):
                # If stdin was a pipe, we need to reopen the terminal for interactive input
                if stdin_was_pipe:
                    try:
                        # Use platform-specific terminal device
                        # Windows: 'CON', Unix/Linux: '/dev/tty'
                        terminal_device = 'CON' if os.name == 'nt' else '/dev/tty'
                        sys.stdin = open(terminal_device, 'r')
                    except (OSError, FileNotFoundError) as e:
                        perror(f"Cannot open terminal for interactive mode: {e}")
                        monitor.stop()
                        self.kills()
                        return True
                print("")
                return False
            # Default behavior: exit after displaying response
            monitor.stop()
            self.kills()
            return True

        return False

    def do_EOF(self, user_content):
        print("")
        return True

    def default(self, user_content):
        # Check for commands before processing multi-line input
        # Normalize: strip whitespace and make case-insensitive
        cmd = user_content.strip().lower()

        # Help command - show available commands
        if cmd in ["/help", "help", "?"]:
            self.do_help("")
            return False

        # Exit commands
        if cmd in ["/bye", "exit"]:
            return True

        # Clear command - reset conversation history and multi-line buffer
        if cmd == "/clear":
            self.conversation_history = []
            self.content = []
            print("Conversation history cleared.")
            return False

        # Handle multi-line input (backslash continuation)
        self.content.append(user_content.removesuffix("\\"))
        if user_content.endswith("\\"):
            return False

        content = "\n".join(self.content)
        self.content = []

        # Check for manual tool selection command FIRST (case-insensitive)
        if self.mcp_agent and content.strip().lower().startswith("/tool"):
            self._handle_manual_tool_selection(content)
            return False

        # Check if MCP agent should handle this request
        if self.mcp_agent and self._should_use_mcp(content):
            response = self._handle_mcp_request(content)
            if response:
                # If streaming, _handle_mcp_request already printed output
                if isinstance(response, str) and response.strip():
                    print(response)
                self.conversation_history.append(UserMessage(text=content))
                self.conversation_history.append(AssistantMessage(text=response))
                self._check_and_summarize()
            return False

        self.conversation_history.append(UserMessage(text=content))
        self.request_in_process = True
        response = self._req()
        if response:
            self.conversation_history.append(AssistantMessage(text=response))
        self.request_in_process = False
        self._check_and_summarize()

    def _make_request_data(self):
        options = self._build_request_options(
            stream=True,
            max_tokens=getattr(self.args, "max_tokens", None),
        )
        request = self.provider.create_request(self.conversation_history, options)
        logger.debug("Request: URL=%s, Data=%s, Headers=%s", request.full_url, request.data, request.headers)
        return request

    def _req(self):
        request = self._make_request_data()

        i: float = 0.01
        total_time_slept: float = 0
        response = None

        max_timeout = 16

        last_error: Optional[Exception] = None

        spinner = Spinner().start()

        while True:
            try:
                response = urllib.request.urlopen(request)
                spinner.stop()
                break
            except urllib.error.HTTPError as http_err:
                error_body = http_err.read().decode("utf-8", "ignore").strip()
                message = f"HTTP {http_err.code}"
                if error_body:
                    message = f"{message}: {error_body}"
                perror(f"\r{message}")

                self.kills()
                spinner.stop()
                return None
            except Exception as exc:
                last_error = exc

            if total_time_slept > max_timeout:
                break

            total_time_slept += i
            time.sleep(i)

            i = min(i * 2, 0.1)

        spinner.stop()
        if response:
            return stream_response(response, self.args.color, self.provider)

        error_suffix = ""
        if last_error:
            error_suffix = f" ({last_error})"
        perror(f"\rError: could not connect to: {self.url}{error_suffix}")
        self.kills()

        return None

    def kills(self):
        # Clean up MCP connections first
        if self.mcp_agent:
            try:
                for client in self.mcp_agent.clients:
                    client.close()
                logger.debug("Closed MCP connections")
            except Exception as e:
                logger.debug(f"Error closing MCP connections: {e}")

        server_process = getattr(self.args, "server_process", None)
        if server_process:
            server_process.terminate()
            try:
                server_process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                server_process.kill()
        elif getattr(self.args, "name", None):
            args = copy.copy(self.args)
            args.ignore = True
            name = getattr(self.args, "name", None)
            # Remove containers on normal exit (remove=True)
            if name:
                stop_container(args, name, remove=True)
            if extra_name := getattr(self.operational_args, "name", None):
                stop_container(args, extra_name, remove=True)

    def loop(self):
        while True:
            self.request_in_process = False
            try:
                self.cmdloop()
            except KeyboardInterrupt:
                # Distinguish between user interrupt and server/container exit
                if self.operational_args.monitor and self.operational_args.monitor.is_exited():
                    # Server/container exited, exit with clear message
                    perror("\nServer or container exited. Shutting down client.")
                    raise
                else:
                    print("")
                    if not self.request_in_process:
                        print("Use /help for commands. Ctrl+d, /bye or exit to quit.")

                continue

            break


class TimeoutException(Exception):
    pass


def alarm_handler(signum, frame):
    """
    Signal handler for SIGALRM. Raises TimeoutException when invoked.
    Note: SIGALRM is Unix-only and not available on Windows.
    """
    raise TimeoutException()


class ServerMonitor:
    """Monitor server process or container and report when it exits."""

    def __init__(
        self,
        server_process=None,
        container_name=None,
        container_engine=None,
        join_timeout=3.0,
        check_interval=0.5,
        inspect_timeout=30.0,
    ):
        """
        Initialize the server monitor.

        Args:
            server_process: subprocess.Popen object to monitor
            container_name: Container name to monitor (for container monitoring)
            container_engine: Container engine command (podman/docker)
            join_timeout: Seconds for thread join when stopping (default: 3.0)
            check_interval: Seconds between monitoring checks (default: 0.5)
            inspect_timeout: Seconds to wait for container inspect command to complete (default: 30.0)

        Note: If neither server_process nor container_name is provided, the monitor
        operates in no-op mode (no actual monitoring occurs).
        """
        self.server_process = server_process
        self.container_name = container_name
        self.container_engine = container_engine
        self.timeout = join_timeout
        self.check_interval = check_interval
        self.inspect_timeout = inspect_timeout
        self._stop_event = threading.Event()
        self._exited_event = threading.Event()
        self._exit_info = {}
        self._monitor_thread = None

        # Determine monitoring mode
        if self.server_process:
            self._mode = "process"
        elif container_name and container_engine:
            self._mode = "container"
        else:
            # No monitoring needed - chat is being used without a service
            self._mode = "none"

    def start(self):
        """Start the monitoring thread."""
        # No-op if not monitoring anything
        if self._mode == "none":
            return

        if self._monitor_thread and self._monitor_thread.is_alive():
            logger.warning("Monitor thread already running")
            return

        if self._mode == "process":
            target = self._monitor_process
        else:
            target = self._monitor_container

        self._monitor_thread = threading.Thread(target=target, daemon=True)
        self._monitor_thread.start()

    def stop(self):
        """Stop the monitoring thread."""
        if self._monitor_thread and self._monitor_thread.is_alive():
            self._stop_event.set()
            self._monitor_thread.join(timeout=self.timeout)

    def is_exited(self):
        """Check if the monitored server/container has exited."""
        return self._exited_event.is_set()

    def is_stopping(self):
        """Check if the monitor is in the process of stopping."""
        return self._stop_event.is_set()

    def get_exit_info(self):
        """Get information about the exit."""
        return self._exit_info.copy()

    def _monitor_process(self):
        """Monitor the server process and report if it exits."""
        while not self._stop_event.is_set():
            try:
                exit_code = self.server_process.poll()
                if exit_code is not None:
                    # Process has exited
                    self._exit_info["pid"] = self.server_process.pid
                    self._exit_info["type"] = "exit"
                    self._exit_info["code"] = exit_code
                    self._exited_event.set()
                    # Send SIGINT to main process to interrupt the chat
                    _thread.interrupt_main()
                    break
            except Exception as e:
                logger.debug(f"Error monitoring process: {e}", exc_info=True)
                self._exit_info["pid"] = self.server_process.pid
                self._exit_info["type"] = "missing"
                self._exited_event.set()
                _thread.interrupt_main()
                break

            # Use wait() instead of sleep() for responsive shutdown
            self._stop_event.wait(self.check_interval)

    def _monitor_container(self):
        """Monitor the container and report if it exits."""
        while not self._stop_event.is_set():
            try:
                # Check if container is still running and get its state and exit code in one go
                inspect_format = "{{.State.Status}}\n{{.State.ExitCode}}"
                result = subprocess.run(
                    [self.container_engine, "inspect", "--format", inspect_format, self.container_name],
                    capture_output=True,
                    text=True,
                    timeout=self.inspect_timeout,
                )
                output_lines = result.stdout.strip().split('\n')
                status = output_lines[0] if output_lines else ""
                exit_code_str = output_lines[1] if len(output_lines) > 1 else ""

                # Explicitly check for non-running states
                if status in ["exited", "dead", "removing"] or (status == "" and exit_code_str != ""):
                    # Container has exited
                    self._exit_info["name"] = self.container_name
                    self._exit_info["type"] = "container"
                    # Default to 'exited' if status is empty but exit code exists
                    self._exit_info["status"] = status if status else "exited"

                    try:
                        self._exit_info["code"] = int(exit_code_str)
                    except (ValueError, AttributeError):
                        self._exit_info["code"] = "unknown"

                    self._exited_event.set()
                    # Send SIGINT to main process to interrupt the chat
                    _thread.interrupt_main()
                    break
            except subprocess.TimeoutExpired:
                logger.debug(f"Timeout checking container {self.container_name} status")
            except subprocess.CalledProcessError:
                # Container not found or error checking status
                self._exit_info["name"] = self.container_name
                self._exit_info["type"] = "container_missing"
                self._exited_event.set()
                _thread.interrupt_main()
                break
            except Exception as e:
                logger.debug(f"Error checking container status: {e}")
            # Use wait() instead of sleep() for responsive shutdown
            self._stop_event.wait(self.check_interval)


def _report_server_exit(monitor):
    """Report details about server exit."""
    exit_info = monitor.get_exit_info()
    exit_type = exit_info.get("type", "unknown")

    if exit_type == "container":
        container_name = exit_info.get("name", "unknown")
        exit_code = exit_info.get("code", "unknown")
        status = exit_info.get("status", "unknown")
        perror(f"Container '{container_name}' exited unexpectedly with exit code {exit_code} (status: {status})")
        perror("\nThe chat session has been terminated because the container is no longer running.")
        perror(f"Check container logs with: {monitor.container_engine} logs {container_name}")
    elif exit_type == "container_missing":
        container_name = exit_info.get("name", "unknown")
        perror(f"Container '{container_name}' not found - may have been removed")
        perror("\nThe chat session has been terminated because the container is no longer available.")
    else:
        # Process-based exit
        pid = exit_info.get("pid", "unknown")

        if exit_type == "exit":
            exit_code = exit_info.get("code", "unknown")
            perror(f"Server process (PID {pid}) exited unexpectedly with exit code {exit_code}")
        elif exit_type == "signal":
            signal_num = exit_info.get("signal", "unknown")
            perror(f"Server process (PID {pid}) was terminated by signal {signal_num}")
        elif exit_type == "missing":
            perror(f"Server process (PID {pid}) not found - may have exited before monitoring started")
        else:
            perror(f"Server process (PID {pid}) exited unexpectedly")

        perror("\nThe chat session has been terminated because the server is no longer running.")
        perror("Check server logs for more details about why the service exited.")


def chat(
    args: ChatArgsType,
    operational_args: Optional[ChatOperationalArgs] = None,
    provider: Optional[ChatProvider] = None,
):
    if args.dryrun:
        assert args.ARGS is not None
        prompt = " ".join(args.ARGS)
        print(f"\nramalama chat --color {args.color} --prefix  \"{args.prefix}\" --url {args.url} {prompt}")
        return

    if provider is None:
        provider = OpenAICompletionsChatProvider(args.url, getattr(args, "api_key", None))

    # SIGALRM is Unix-only, skip keepalive timeout handling on Windows
    if getattr(args, "keepalive", False) and hasattr(signal, 'SIGALRM'):
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(convert_to_seconds(args.keepalive))  # type: ignore

    # Start server process or container monitoring
    server_process = getattr(args, "server_process", None)
    container_name = getattr(args, "name", None)

    if server_process:
        # Monitor the server process
        monitor = ServerMonitor(server_process=server_process)
    elif container_name:
        # Monitor the container
        conman = getattr(args, "engine", ActiveConfig().engine)
        if not conman:
            raise ValueError("Container engine is required when monitoring a container")
        monitor = ServerMonitor(container_name=container_name, container_engine=conman)
    else:
        # No monitoring needed - chat is being used directly without a service
        monitor = ServerMonitor()

    monitor.start()
    list_models = getattr(args, "list", False)
    if list_models:
        for model_id in provider.list_models():
            print(model_id)
        monitor.stop()
        if hasattr(signal, 'alarm'):
            signal.alarm(0)
        return

    # Ensure operational_args is initialized
    if operational_args is None:
        operational_args = ChatOperationalArgs()

    # Assign monitor to operational_args
    operational_args.monitor = monitor

    successful_exit = True
    try:
        shell = RamaLamaShell(args, operational_args, provider=provider)
        if shell.handle_args(monitor):
            return

        if not list_models:
            shell.loop()
    except KeyboardInterrupt:
        # If monitor was already shut down exit cleanly
        if monitor.is_stopping():
            return
        # Check if this was caused by server exit
        if monitor.is_exited():
            successful_exit = False  # Server/container exited unexpectedly
            perror("\n" + "=" * 60)
            _report_server_exit(monitor)
            perror("=" * 60)
            exit_info = monitor.get_exit_info()
            exit_code = exit_info.get('code', 1)
            if exit_info.get('type') == 'signal' and isinstance(exit_info.get('signal'), int):
                exit_code = 128 + exit_info.get('signal')
            raise SystemExit(exit_code)
        raise
    except TimeoutException as e:
        logger.debug(f"Timeout Exception: {e}")
        # Handle the timeout, e.g., print a message and exit gracefully
        perror("")
        pass
    finally:
        # Stop the monitoring thread
        monitor.stop()
        # Reset the alarm to 0 to cancel any pending alarms (Unix only)
        if hasattr(signal, 'alarm'):
            signal.alarm(0)

    # Only clean up resources on successful exit
    # If server/container crashed, leave it for log inspection
    if successful_exit:
        try:
            shell.kills()
        except Exception as e:
            logger.warning(f"Failed to clean up resources: {e}")


UNITS = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days", "w": "weeks"}


def convert_to_seconds(s):
    if isinstance(s, int):
        # We are dealing with a raw number
        return s

    try:
        seconds = int(s)
        # We are dealing with an integer string
        return seconds
    except ValueError:
        # We are dealing with some other string or type
        pass

    # Expecting a string ending in [m|h|d|s|w]
    count = int(s[:-1])
    unit = UNITS[s[-1]]
    td = timedelta(**{unit: count})
    return td.seconds + 60 * 60 * 24 * td.days



from __future__ import annotations

import logging
from enum import IntEnum
from typing import Union


class LogLevel(IntEnum):
    DEBUG = getattr(logging, "DEBUG")
    INFO = getattr(logging, "INFO")
    WARNING = getattr(logging, "WARNING")
    ERROR = getattr(logging, "ERROR")
    CRITICAL = getattr(logging, "CRITICAL")


def coerce_log_level(level: Union["LogLevel", str, int]) -> LogLevel:
    if isinstance(level, LogLevel):
        return level
    if isinstance(level, str):
        try:
            return LogLevel[level.upper()]
        except KeyError as exc:
            raise ValueError(f"Unsupported log level: {level}") from exc
    if isinstance(level, int):
        try:
            return LogLevel(level)
        except ValueError as exc:
            raise ValueError(f"Unsupported log level value: {level}") from exc
    raise TypeError(f"Cannot coerce {level!r} to LogLevel")



from __future__ import annotations

from collections.abc import Callable
from typing import Optional

from ramalama.chat_providers.base import ChatProvider
from ramalama.chat_providers.openai import OpenAIResponsesChatProvider
from ramalama.config import ActiveConfig

PROVIDER_API_KEY_RESOLVERS: dict[str, Callable[[], Optional[str]]] = {
    "openai": lambda: ActiveConfig().provider.openai.api_key,
}


def get_provider_api_key(scheme: str) -> Optional[str]:
    """Return a configured API key for the given provider scheme, if any."""

    if resolver := PROVIDER_API_KEY_RESOLVERS.get(scheme):
        if key := resolver():
            return key
    return ActiveConfig().api_key


DEFAULT_PROVIDERS = {
    "openai": lambda: OpenAIResponsesChatProvider(
        base_url="https://api.openai.com/v1", api_key=get_provider_api_key("openai")
    )
}


def get_chat_provider(scheme: str) -> ChatProvider:
    if (resolver := DEFAULT_PROVIDERS.get(scheme, None)) is None:
        raise ValueError(f"No supported chat provider for {scheme}")
    return resolver()


__all__ = ["get_chat_provider", "get_provider_api_key"]



from __future__ import annotations

import os
import shutil
import urllib.error
from collections import Counter
from http import HTTPStatus
from pathlib import Path
from typing import Optional, Sequence, Tuple

from ramalama.common import perror, sanitize_filename, verify_checksum
from ramalama.endian import EndianMismatchError, get_system_endianness
from ramalama.logger import logger
from ramalama.model_inspect.gguf_parser import GGUFInfoParser, GGUFModelInfo
from ramalama.model_store import go2jinja
from ramalama.model_store.constants import DIRECTORY_NAME_BLOBS, DIRECTORY_NAME_REFS, DIRECTORY_NAME_SNAPSHOTS
from ramalama.model_store.global_store import GlobalModelStore
from ramalama.model_store.reffile import RefJSONFile, StoreFile, StoreFileType, migrate_reffile_to_refjsonfile
from ramalama.model_store.snapshot_file import (
    LocalSnapshotFile,
    SnapshotFile,
    SnapshotFileType,
    validate_snapshot_files,
)
from ramalama.model_store.template_conversion import (
    TemplateConversionError,
    convert_go_to_jinja,
    ensure_jinja_openai_compatibility,
    is_openai_jinja,
)
from ramalama.path_utils import create_file_link


def map_to_store_file_type(snapshot_type: SnapshotFileType) -> StoreFileType:
    mapping = {
        SnapshotFileType.GGUFModel: StoreFileType.GGUF_MODEL,
        SnapshotFileType.ChatTemplate: StoreFileType.CHAT_TEMPLATE,
        SnapshotFileType.Mmproj: StoreFileType.MMPROJ,
        SnapshotFileType.SafetensorModel: StoreFileType.SAFETENSOR_MODEL,
    }
    return mapping.get(snapshot_type, StoreFileType.OTHER)


class ModelStore:
    def __init__(
        self,
        store: GlobalModelStore,
        model_name: str,
        model_type: str,
        model_organization: str,
    ):
        self._store = store
        self._model_name = model_name
        self._model_type = model_type
        self._model_organization = model_organization

    @property
    def base_path(self) -> str:
        return self._store.path

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def model_organization(self) -> str:
        return self._model_organization if self._model_organization != "" else self._model_name

    @property
    def model_type(self) -> str:
        return self._model_type

    @property
    def model_base_directory(self) -> str:
        return os.path.join(self.base_path, self.model_type, self.model_organization, self.model_name)

    @property
    def blobs_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_BLOBS)

    @property
    def refs_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_REFS)

    @property
    def snapshots_directory(self) -> str:
        return os.path.join(self.model_base_directory, DIRECTORY_NAME_SNAPSHOTS)

    def file_exists(self, file_path: str) -> bool:
        return os.path.exists(file_path)

    def get_ref_file_path(self, model_tag: str) -> str:
        return os.path.join(self.refs_directory, f"{model_tag}.json")

    def get_ref_file(self, model_tag: str) -> Optional[RefJSONFile]:
        ref_file_path = self.get_ref_file_path(model_tag)
        ref_file = migrate_reffile_to_refjsonfile(ref_file_path, self.snapshots_directory)
        if ref_file is None:
            if os.path.exists(ref_file_path):
                ref_file = RefJSONFile.from_path(ref_file_path)
        if ref_file is not None:
            if ref_file.version != RefJSONFile.version:
                # 0.13.0 chat template conversion logic was wrong, force a refresh
                ref_file.version = RefJSONFile.version
                self._ensure_chat_template(ref_file, snapshot_hash=ref_file.hash)
        return ref_file

    def update_ref_file(
        self, model_tag: str, snapshot_hash: str = "", snapshot_files: Optional[list[SnapshotFile]] = None
    ) -> Optional[RefJSONFile]:
        if snapshot_files is None:
            snapshot_files = []

        ref_file: Optional[RefJSONFile] = self.get_ref_file(model_tag)
        if ref_file is None:
            return None

        if snapshot_hash != "":
            ref_file.hash = snapshot_hash
        if snapshot_files != []:
            ref_file.files = []
        for file in snapshot_files:
            ref_file.files.append(StoreFile(file.hash, file.name, map_to_store_file_type(file.type)))

        ref_file.write_to_file()

        return ref_file

    def get_snapshot_hash(self, model_tag: str) -> str:
        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            return ""
        return sanitize_filename(ref_file.hash)

    def get_snapshot_directory_from_tag(self, model_tag: str) -> str:
        return os.path.join(self.snapshots_directory, self.get_snapshot_hash(model_tag))

    def get_snapshot_directory(self, hash: str) -> str:
        return os.path.join(self.snapshots_directory, hash)

    def get_snapshot_file_path(self, tag_hash: str, filename: str) -> str:
        return os.path.join(self.snapshots_directory, sanitize_filename(tag_hash), filename)

    def get_blob_file_path(self, file_hash: str) -> str:
        return os.path.join(self.blobs_directory, sanitize_filename(file_hash))

    def get_safetensor_blob_path(self, model_tag: str, requested_filename: str) -> Optional[str]:
        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            return None
        safetensor_files = ref_file.safetensor_model_files
        if not safetensor_files:
            return None
        matched = next((f for f in safetensor_files if f.name == requested_filename), None)
        chosen = matched if matched is not None else safetensor_files[0]
        return self.get_blob_file_path(chosen.hash)

    def get_blob_file_path_by_name(self, tag_hash: str, filename: str) -> str:
        return str(Path(self.get_snapshot_file_path(tag_hash, filename)).resolve())

    def get_blob_file_hash(self, tag_hash: str, filename: str) -> str:
        return os.path.basename(self.get_blob_file_path_by_name(tag_hash, filename))

    def get_partial_blob_file_path(self, file_hash: str) -> str:
        return self.get_blob_file_path(file_hash) + ".partial"

    def ensure_directory_setup(self) -> None:
        os.makedirs(self.blobs_directory, exist_ok=True)
        os.makedirs(self.refs_directory, exist_ok=True)
        os.makedirs(self.snapshots_directory, exist_ok=True)

    def directory_setup_exists(self) -> bool:
        return (
            os.path.exists(self.blobs_directory)
            and os.path.exists(self.refs_directory)
            and os.path.exists(self.snapshots_directory)
        )

    def get_cached_files(self, model_tag: str) -> Tuple[str, list[str], bool]:
        cached_files: list[str] = []

        ref_file: Optional[RefJSONFile] = self.get_ref_file(model_tag)
        if ref_file is None:
            return ("", cached_files, False)

        # TODO: Remove in following releases
        # Temporary migration of .safetensors model files which were previously stored as OTHER
        should_write = False
        for file in ref_file.files:
            if file.name.endswith(".safetensors") and file.type != StoreFileType.SAFETENSOR_MODEL:
                file.type = StoreFileType.SAFETENSOR_MODEL
                should_write = True
        if should_write:
            ref_file.write_to_file()

        for file in ref_file.files:
            path = self.get_blob_file_path(file.hash)
            if os.path.exists(path):
                cached_files.append(file.name)

        return (ref_file.hash, cached_files, len(cached_files) == len(ref_file.files))

    def _prepare_new_snapshot(
        self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]
    ) -> RefJSONFile:
        validate_snapshot_files(snapshot_files)
        self.ensure_directory_setup()

        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            ref_file = RefJSONFile(snapshot_hash, self.get_ref_file_path(model_tag), [])
            for file in snapshot_files:
                ref_file.files.append(StoreFile(file.hash, file.name, map_to_store_file_type(file.type)))

            ref_file.write_to_file()

        snapshot_directory = self.get_snapshot_directory(snapshot_hash)
        os.makedirs(snapshot_directory, exist_ok=True)
        return ref_file

    def _download_snapshot_files(
        self, ref_file: RefJSONFile, snapshot_hash: str, snapshot_files: Sequence[SnapshotFile]
    ):
        for file in snapshot_files:
            dest_path = self.get_blob_file_path(file.hash)
            try:
                file.download(dest_path, self.get_snapshot_directory(snapshot_hash))
            except urllib.error.HTTPError as ex:
                if file.required:
                    raise ex
                # remove file from ref file list to prevent a retry to download it
                if ex.code == HTTPStatus.NOT_FOUND:
                    ref_file.remove_file(file.hash)
                continue

            if file.should_verify_checksum:
                if not verify_checksum(dest_path):
                    logger.info(f"Checksum mismatch for blob {dest_path}, retrying download ...")
                    os.remove(dest_path)
                    file.download(dest_path, self.get_snapshot_directory(snapshot_hash))
                    if not verify_checksum(dest_path):
                        raise ValueError(f"Checksum verification failed for blob {dest_path}")

            link_path = self.get_snapshot_file_path(snapshot_hash, file.name)

            blob_absolute_path = self.get_blob_file_path(file.hash)
            # Use cross-platform file linking (hardlink/symlink/copy)
            create_file_link(blob_absolute_path, link_path)

        # save updated ref file
        ref_file.write_to_file()

    def _try_convert_existing_chat_template(self, ref_file: RefJSONFile, snapshot_hash: str) -> bool:
        for file in ref_file.chat_templates:
            chat_template_file_path = self.get_blob_file_path(file.hash)
            with open(chat_template_file_path, "r") as template_file:
                chat_template = template_file.read()

            if not go2jinja.is_go_template(chat_template):
                if is_openai_jinja(chat_template):
                    return True
                else:
                    normalized_template = ensure_jinja_openai_compatibility(chat_template)
            else:
                try:
                    normalized_template = convert_go_to_jinja(chat_template)
                except TemplateConversionError as e:
                    logger.debug(f"Failed to convert template: {e}")
                    continue

            files = [
                LocalSnapshotFile(
                    normalized_template.encode("utf-8"), "chat_template_converted", SnapshotFileType.ChatTemplate
                )
            ]
            self._update_snapshot(ref_file, snapshot_hash, files)
            return True

        return False

    def _ensure_chat_template(self, ref_file: RefJSONFile, snapshot_hash: str):
        # Give preference to the embedded chat template as it's most likely to be
        # compatible with llama.cpp

        def get_embedded_template() -> Optional[str]:
            models = ref_file.model_files
            if not models:
                return None

            # Only the first model file is considered for chat template extraction
            model_file_path = self.get_blob_file_path(models[0].hash)
            if not GGUFInfoParser.is_model_gguf(model_file_path):
                return None

            # Parse model, first and second parameter are irrelevant here
            info: GGUFModelInfo = GGUFInfoParser.parse("model", "registry", model_file_path)
            return info.get_chat_template()

        tmpl = get_embedded_template()

        if tmpl is None:
            self._try_convert_existing_chat_template(ref_file, snapshot_hash)
            return

        needs_conversion = go2jinja.is_go_template(tmpl)

        # Only jinja templates are usable for the supported backends, therefore don't mark file as
        # chat template if it is a Go Template (ollama-specific)
        files = [
            LocalSnapshotFile(
                tmpl.encode("utf-8"),
                "chat_template_extracted",
                SnapshotFileType.Other if needs_conversion else SnapshotFileType.ChatTemplate,
            )
        ]
        if needs_conversion:
            try:
                desired_template = convert_go_to_jinja(tmpl)
                files.append(
                    LocalSnapshotFile(
                        desired_template.encode("utf-8"), "chat_template_converted", SnapshotFileType.ChatTemplate
                    )
                )
            except Exception as ex:
                logger.debug(f"Failed to convert Go Template to Jinja: {ex}")
        else:
            for file in ref_file.files:
                if file.name == "chat_template_converted":
                    # Should not exist but 0.13.0 needs_conversion logic was inverted
                    self._remove_blob_path(Path(self.get_snapshot_file_path(ref_file.hash, file.name)))
                    ref_file.remove_file(file.hash)
                    break
        self._update_snapshot(ref_file, snapshot_hash, files)

    def _verify_endianness(self, model_tag: str):
        ref_file = self.get_ref_file(model_tag)
        if ref_file is None:
            return

        for model_file in ref_file.model_files:
            model_path = self.get_blob_file_path(model_file.hash)

            # only check endianness for gguf models
            if not GGUFInfoParser.is_model_gguf(model_path):
                return

            model_endianness = GGUFInfoParser.get_model_endianness(model_path)
            host_endianness = get_system_endianness()
            if host_endianness != model_endianness:
                raise EndianMismatchError(host_endianness, model_endianness)

    def verify_snapshot(self, model_tag: str):
        self._verify_endianness(model_tag)
        self._store.verify_snapshot()

    def new_snapshot(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile], verify: bool = True):
        snapshot_hash = sanitize_filename(snapshot_hash)

        try:
            ref_file = self._prepare_new_snapshot(model_tag, snapshot_hash, snapshot_files)
            self._download_snapshot_files(ref_file, snapshot_hash, snapshot_files)
            self._ensure_chat_template(ref_file, snapshot_hash)
        except urllib.error.HTTPError as ex:
            perror(f"Failed to fetch required file: {ex}")
            perror("Removing snapshot...")
            self.remove_snapshot(model_tag)
            raise ex
        except Exception as ex:
            perror("Removing snapshot...")
            self.remove_snapshot(model_tag)
            raise ex

        try:
            if verify:
                self.verify_snapshot(model_tag)
        except EndianMismatchError as ex:
            perror(f"Verification of snapshot failed: {ex}")
            perror("Removing snapshot...")
            self.remove_snapshot(model_tag)
            raise ex

    def _update_snapshot(
        self, ref_file: RefJSONFile, snapshot_hash: str, new_snapshot_files: Sequence[SnapshotFile]
    ) -> bool:
        validate_snapshot_files(new_snapshot_files)
        snapshot_hash = sanitize_filename(snapshot_hash)

        if not self.directory_setup_exists():
            return False

        # update ref file with deduplication by file hash
        existing_file_hashes = {f.hash for f in ref_file.files}
        for new_snapshot_file in new_snapshot_files:
            if new_snapshot_file.hash not in existing_file_hashes:
                ref_file.files.append(
                    StoreFile(
                        new_snapshot_file.hash,
                        new_snapshot_file.name,
                        map_to_store_file_type(new_snapshot_file.type),
                    )
                )
        ref_file.write_to_file()

        self._download_snapshot_files(ref_file, snapshot_hash, new_snapshot_files)
        return True

    def _remove_blob_path(self, blob_path: Path):
        try:
            if blob_path.exists() and Path(self.base_path) in blob_path.parents:
                blob_path.unlink(missing_ok=True)
                logger.debug(f"Removed blob file '{blob_path}'")
        except Exception as ex:
            logger.error(f"Failed to remove blob file '{blob_path}': {ex}")

    def _get_refcounts(self, snapshot_hash: str) -> tuple[int, Counter[str]]:
        # get all ref file names and remove the last suffix, i.e. .json, if it exists
        # so that only the model tag remains
        model_tags = [
            Path(entry).stem
            for entry in os.listdir(self.refs_directory)
            if os.path.isfile(os.path.join(self.refs_directory, entry))
        ]
        refs = [ref for tag in model_tags if (ref := self.get_ref_file(tag))]

        blob_refcounts = Counter(file.name for ref in refs for file in ref.files)

        snap_refcount = sum(ref.hash == snapshot_hash for ref in refs)

        return snap_refcount, blob_refcounts

    def remove_snapshot(self, model_tag: str) -> bool:
        ref_file = self.get_ref_file(model_tag)

        if ref_file is None:
            return False

        snapshot_refcount, blob_refcounts = self._get_refcounts(ref_file.hash)

        # Remove all blobs first
        for file in ref_file.files:
            blob_refcount = blob_refcounts.get(file.name, 0)
            if blob_refcount <= 1:
                blob_absolute_path = Path(self.get_blob_file_path(file.hash))
                self._remove_blob_path(blob_absolute_path)
            else:
                logger.debug(f"Not removing blob {file} refcount={blob_refcount}")

        # Remove snapshot directory
        if snapshot_refcount <= 1:
            # FIXME: this only cleans up .partial files where the blob hash equals the snapshot hash
            partial_blob_file_path = Path(self.get_partial_blob_file_path(ref_file.hash))
            self._remove_blob_path(partial_blob_file_path)
            snapshot_directory = self.get_snapshot_directory_from_tag(model_tag)
            shutil.rmtree(snapshot_directory, ignore_errors=True)
            logger.debug(f"Snapshot removed {ref_file.hash}")
        else:
            logger.debug(f"Not removing snapshot {ref_file.hash} refcount={snapshot_refcount}")

        # Remove ref file, ignore if file is not found
        Path(self.get_ref_file_path(model_tag)).unlink(missing_ok=True)
        return True



from __future__ import annotations

import sys
from enum import IntEnum


class GGUFEndian(IntEnum):
    LITTLE = 0
    BIG = 1

    def __str__(self):
        return self.name


def get_system_endianness() -> GGUFEndian:
    return GGUFEndian.LITTLE if sys.byteorder == 'little' else GGUFEndian.BIG


class EndianMismatchError(Exception):
    def __init__(self, host_endianness: GGUFEndian, model_endianness: GGUFEndian, *args):
        super().__init__(f"Endian mismatch of host ({host_endianness}) and model ({model_endianness})", *args)



from __future__ import annotations

import os
import subprocess
from typing import Optional

from ramalama.common import MNT_DIR, run_cmd
from ramalama.transports.oci.oci import OCI


def find_model_file_in_image(conman: str, model: str) -> Optional[str]:
    """Inspect the OCI image to find model file location"""
    # First try to get from label
    try:
        cmd = [
            conman,
            "image",
            "inspect",
            "--format={{index .Config.Labels \"com.ramalama.model.file.location\"}}",
            model,
        ]
        result = run_cmd(cmd)
        model_file = result.stdout.decode('utf-8').strip()

        if model_file and model_file != '<no value>':
            return model_file
    except subprocess.CalledProcessError:
        pass

    # Fallback: try to find .gguf files in /models directory
    try:
        cmd = [conman, "run", "--rm", model, "ls", "/models"]
        result = run_cmd(cmd)
        files = result.stdout.decode('utf-8').strip().split('\n')

        # Look for .gguf files
        gguf_files = [f for f in files if f.endswith('.gguf')]
        if gguf_files:
            return gguf_files[0]  # Return first .gguf file found
    except subprocess.CalledProcessError:
        pass

    return None


class RamalamaContainerRegistry(OCI):
    def __init__(self, model: str, *args, **kwargs):
        super().__init__(f"rlcr.io/ramalama/{model}", *args, **kwargs)
        self._model_type = "oci"

    def _get_entry_model_path(self, use_container: bool, should_generate: bool, dry_run: bool) -> str:
        if dry_run:
            return super()._get_entry_model_path(use_container, should_generate, dry_run)

        if self.strategy.kind == 'artifact':
            return super()._get_entry_model_path(use_container, should_generate, dry_run)

        if use_container or should_generate:
            model_filename = find_model_file_in_image(self.conman, self.model)
            if model_filename:
                model_basename = os.path.basename(model_filename)
                return os.path.join(MNT_DIR, model_basename)

        return super()._get_entry_model_path(use_container, should_generate, dry_run)



from __future__ import annotations

import os
import subprocess
import tempfile
from functools import partial
from textwrap import dedent
from typing import Literal

from ramalama.arg_types import RagArgsType
from ramalama.chat import ChatOperationalArgs
from ramalama.common import ensure_image, perror, set_accel_env_vars
from ramalama.compat import StrEnum
from ramalama.config import ActiveConfig, Config
from ramalama.engine import BuildEngine, Engine, is_healthy, stop_container, wait_for_healthy
from ramalama.path_utils import get_container_mount_path
from ramalama.transports.base import Transport
from ramalama.transports.oci.oci import OCI

INPUT_DIR = "/docs"


class VectorDBEngine(Engine):
    """Generate a RAG vector database from source content."""

    def __init__(self, args):
        super().__init__(args)
        self.add_user()

    def add_input(self, path: str) -> None:
        if os.path.exists(path):
            fpath = os.path.realpath(path)
            input_name = os.path.basename(fpath)
            self.add_volume(fpath, f"{INPUT_DIR}/{input_name}")
        else:
            raise ValueError(f"{path} does not exist")

    def add_user(self):
        if self.use_docker:
            # Run the command in the container with the same uid as the current user.
            # Otherwise the files written to the tempdir will be owned by root and
            # they will not be readable by the subsequent "docker build".
            # Note: geteuid() doesn't exist on Windows, but this is only needed on Unix
            if hasattr(os, 'geteuid'):
                self.add_args("--user", str(os.geteuid()))


class Rag:
    target: str = ""
    oci: bool = False

    def __init__(self, target: str):
        self.target = target
        self.oci = self.target.startswith("oci://") or (
            not self.target.startswith("file://") and self.target[0] not in {".", "/"}
        )
        if self.oci and not self.target.islower():
            raise ValueError(f"invalid reference format: repository name '{self.target}' must be lowercase")
        set_accel_env_vars()

    def build(self, source: str, args):
        perror(f"Building {self.target} ...")
        contextdir = os.path.dirname(source)
        src = os.path.basename(source)
        engine = BuildEngine(args)
        return engine.build_containerfile(
            dedent(f"""
                FROM scratch
                COPY {src} /vector.db
                """),
            contextdir,
            tag=self.target,
        )

    def generate(self, args, cmd):
        args.nocapdrop = True
        if not args.container:
            raise KeyError("rag command requires a container. Can not be run with --nocontainer option.")
        if not args.engine or args.engine == "":
            raise KeyError("rag command requires a container. Can not be run without a container engine.")
        engine = VectorDBEngine(args)

        for path in args.PATHS:
            engine.add_input(path)

        if self.oci:
            ragdb = tempfile.TemporaryDirectory(prefix="RamaLama_rag_")
            dbdir = os.path.join(ragdb.name, "vectordb")
            os.makedirs(dbdir)
        else:
            dbdir = self.target

        engine.add_volume(dbdir, "/output", opts="rw")
        if not args.dryrun:
            config = ActiveConfig()
            should_pull = config.pull in ["always", "missing", "newer"]
            args.image = ensure_image(
                args.engine, args.image, should_pull=should_pull, quiet=getattr(args, "quiet", False)
            )
        engine.add_args(args.image)
        engine.add_args(*cmd)
        try:
            if args.dryrun:
                engine.dryrun()
            else:
                engine.run()
            if self.oci:
                print(self.build(dbdir, args))
        except subprocess.CalledProcessError as e:
            raise e
        finally:
            if self.oci:
                ragdb.cleanup()


def rag_image(config: Config) -> str:
    return config.default_rag_image


class RagSource(StrEnum):
    DB = "db"
    IMAGE = "image"


class RagEngine(Engine):
    """Engine for executing a RAG proxy"""

    sourcetype: RagSource

    def __init__(self, args, /, *, sourcetype: RagSource):
        self.sourcetype = sourcetype
        super().__init__(args)
        self.add_rag()

    def add_labels(self):
        super().add_labels()
        rag = getattr(self.args, "rag", None)
        if rag is not None:
            self.add_label(f"ai.ramalama.rag.{self.sourcetype}={rag}")

    def add_rag(self):
        rag = getattr(self.args, "rag", None)
        if rag is None:
            return
        if self.sourcetype is RagSource.DB:
            # Convert to container-friendly path format (handles Windows path conversion)
            rag_path = get_container_mount_path(rag)
            # Read-write is the default behaviour for bind mounts in both Docker and Podman
            self.add_args(f"--mount=type=bind,source={rag_path},destination=/rag/vector.db{self.relabel()}")
        else:
            # Image mounts default to read-only in Podman, so we need rw=true for write access
            # Docker does not support type=image mounts
            if self.use_podman:
                self.add_args(f"--mount=type=image,source={rag},destination=/rag,rw=true")
            else:
                # Docker falls back to using volumes or other mechanisms
                self.add_args(f"--mount=type=image,source={rag},destination=/rag")


class RagTransport(OCI):
    """Run a RAG proxy, dispatching to a backend LLM"""

    type: str = "Model+RAG"

    def format_model(self, model: str) -> str:
        if getattr(self, "kind", None) is RagSource.DB:
            return model
        if "/" not in model:
            model = f"localhost/{model}"
        return super().format_model(model)

    def __init__(self, imodel: Transport, cmd: list[str], args: RagArgsType):
        self.kind = RagSource.DB if os.path.exists(args.rag) else RagSource.IMAGE
        super().__init__(args.rag, args.store, args.engine)
        args.rag = self.model
        self.imodel = imodel
        self.model_cmd = cmd

    def exists(self) -> bool:
        if self.kind is RagSource.DB:
            return os.path.exists(self.model)
        return super().exists()

    def new_engine(self, args: RagArgsType) -> "RagEngine":
        return RagEngine(args, sourcetype=self.kind)

    def setup_mounts(self, args: RagArgsType) -> None:
        pass

    def chat_operational_args(self, args: RagArgsType) -> ChatOperationalArgs:
        return ChatOperationalArgs(name=args.model_args.name)

    def _handle_container_chat(self, args: RagArgsType, server_process: int) -> Literal[0]:
        # Clear args.rag so RamaLamaShell doesn't treat it as local data for RAG context
        args.rag = None  # type: ignore[assignment]
        return super()._handle_container_chat(args, server_process)

    def serve(self, args: RagArgsType, cmd: list[str]):
        args.model_args.name = self.imodel.get_container_name(args.model_args)
        process = self.imodel.serve_nonblocking(args.model_args, self.model_cmd)
        if not args.dryrun:
            if process and process.wait() != 0:
                raise subprocess.CalledProcessError(
                    process.returncode,
                    " ".join(self.model_cmd),
                )
        try:
            self.execute_command(cmd, args)
        finally:
            if getattr(args.model_args, "name", None):
                args.model_args.ignore = True
                stop_container(args.model_args, args.model_args.name, remove=True)

    def run(self, args: RagArgsType, cmd: list[str]):
        args.model_args.name = self.imodel.get_container_name(args.model_args)
        process = self.imodel.serve_nonblocking(args.model_args, self.model_cmd)
        rag_process = self.serve_nonblocking(args, cmd)

        if args.dryrun:
            return

        if process and process.wait() != 0:
            raise subprocess.CalledProcessError(
                process.returncode,
                " ".join(self.model_cmd),
            )
        if rag_process and rag_process.wait() != 0:
            raise subprocess.CalledProcessError(
                rag_process.returncode,
                " ".join(cmd),
            )
        return self._connect_and_chat(args, rag_process)

    def wait_for_healthy(self, args: RagArgsType) -> None:
        self.imodel.wait_for_healthy(args.model_args)
        wait_for_healthy(args, partial(is_healthy, model_name=f"{self.imodel.model_alias}+rag"))



"""Utilities for normalizing and formatting bind host addresses."""

from __future__ import annotations

from typing import Optional

WILDCARD_BIND_HOSTS = frozenset({"0.0.0.0", "::"})


def normalize_bind_host(host: Optional[str]) -> str:
    """Remove surrounding brackets from a bind host value."""
    if not host:
        return ""
    return host.strip("[]")


def localhost_from_bind_host(host: Optional[str], *, loopback: str = "127.0.0.1") -> str:
    """Return loopback when host binds all interfaces, otherwise return the host."""
    if not host:
        return loopback
    normalized = normalize_bind_host(host)
    return loopback if normalized in WILDCARD_BIND_HOSTS else normalized


def format_bind_host_for_url(host: Optional[str]) -> str:
    """Format a bind host for use in an http URL authority (handles IPv6 bracketing)."""
    if not host:
        return "127.0.0.1"
    host = localhost_from_bind_host(host)
    return f"[{host}]" if ":" in host else host


def format_bind_host_for_connection(host: Optional[str], *, default: str = "127.0.0.1") -> str:
    """Format a bind host for HTTPConnection and similar clients."""
    if not host:
        return default
    return localhost_from_bind_host(host, loopback=default)


def format_bind_host_literal(host: Optional[str]) -> str:
    """Format a bind host for display or host:port binding without loopback substitution."""
    if not host:
        return ""
    normalized = normalize_bind_host(host)
    return f"[{normalized}]" if ":" in normalized else normalized


def format_bind_host_publish_prefix(host: Optional[str]) -> str:
    """Return host prefix for container publish-port arguments, or empty for IPv6 wildcard."""
    if not host:
        return ""
    normalized = normalize_bind_host(host)
    if normalized == "::":
        return ""
    return f"[{normalized}]:" if ":" in normalized else f"{normalized}:"



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import errno
import signal
import socket
import socketserver
import threading
from datetime import datetime, timedelta
from typing import Optional

from ramalama.config import ActiveConfig
from ramalama.daemon.handler.ramalama import RamalamaHandler
from ramalama.daemon.logging import configure_logger, logger
from ramalama.daemon.service.model_runner import ModelRunner
from ramalama.host_utils import format_bind_host_literal
from ramalama.log_levels import LogLevel


class ShutdownHandler:
    def __init__(self, server: "RamalamaServer") -> None:
        self.server = server
        self.idle_check_timer: Optional[threading.Timer] = None

    def handle_kill(self, signum, frame):
        if self.idle_check_timer:
            self.idle_check_timer.cancel()
        self.server.shutdown()

    def handle_alarm(self, signum, frame):
        # check for expiration of all models, stopping them if necessary and
        # stop shutdown server if no models are running
        self.server.check_model_expiration()
        if not self.server.model_runner.managed_models:
            self.server.shutdown()
            return

        # register alarm again for next check (Unix only)
        if hasattr(signal, 'alarm'):
            signal.alarm(self.server.idle_check_interval.seconds)

    def __enter__(self):
        signal.signal(signal.SIGINT, self.handle_kill)
        signal.signal(signal.SIGTERM, self.handle_kill)

        # SIGALRM is Unix-only, use threading.Timer on Windows as fallback
        if hasattr(signal, 'SIGALRM'):
            # set initial idle check to 300s == 5min to prevent service from stopping
            # right after being started
            signal.signal(signal.SIGALRM, self.handle_alarm)
            signal.alarm(300)
        else:
            # On Windows, use a background thread for idle checks
            self._start_idle_check_thread()

    def _start_idle_check_thread(self, interval_seconds=300):
        """Start a background thread for idle checking on Windows (no SIGALRM support)."""

        def check_and_reschedule():
            self.server.check_model_expiration()
            if not self.server.model_runner.managed_models:
                self.server.shutdown()
                return
            # Schedule next check
            self.idle_check_timer = threading.Timer(self.server.idle_check_interval.seconds, check_and_reschedule)
            self.idle_check_timer.daemon = True
            self.idle_check_timer.start()

        self.idle_check_timer = threading.Timer(interval_seconds, check_and_reschedule)
        self.idle_check_timer.daemon = True
        self.idle_check_timer.start()

    def __exit__(self, type, value, traceback):
        if self.idle_check_timer is not None:
            self.idle_check_timer.cancel()


class RamalamaServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    allow_reuse_address = True

    def __init__(
        self, host: str, port: int, model_store_path: str, idle_check_interval: timedelta, bind_and_activate=True
    ):
        # Use AF_INET6 for IPv6 addresses; on dual-stack systems :: accepts IPv4 too
        if ":" in host:
            self.address_family = socket.AF_INET6
        # Do not pass a RequestHandlerClass here, we will create a custom handler in finish_request
        super().__init__((host, port), None, bind_and_activate)  # type: ignore

        self.model_store_path: str = model_store_path
        self.model_runner: ModelRunner = ModelRunner()
        self.idle_check_interval: timedelta = idle_check_interval

    def server_bind(self):
        # Enable dual-stack so :: accepts IPv4 connections too
        if self.address_family == socket.AF_INET6:
            try:
                self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            except (AttributeError, OSError):
                pass
        super().server_bind()

    def finish_request(self, request, client_address):
        RamalamaHandler(self.model_store_path, self.model_runner, request, client_address, self)

    def check_model_expiration(self):
        curr_time = datetime.now()
        for name, m in self.model_runner.managed_models.items():
            expiration_date = getattr(m, "expiration_date", None)
            if expiration_date is None or expiration_date > curr_time:
                continue

            try:
                logger.info(f"Stopping expired model '{name}'...")
                self.model_runner.stop_model(m.id)
            except Exception as e:
                logger.error(f"Failed to stop expired model '{name}': {e}")

    def shutdown(self):
        logger.info("Shutting down ramalama daemon...")

        for name, managed_model in self.model_runner.managed_models.items():
            try:
                logger.info(f"Stopping model runner {name}...")
                self.model_runner.stop_model(managed_model.id)
            except Exception as e:
                logger.error(f"Error stopping model runner {name}: {e}")

        super().shutdown()


def parse_args():
    parser = argparse.ArgumentParser(description="Ramalama Daemon")
    parser.add_argument("--host", type=str, default="::")
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument("--model-store-path", type=str, default="/models")

    return parser.parse_args()


def run(host: str = "::", port: int = 8080, model_store_path: str = "/models"):
    configure_logger(ActiveConfig().log_level or LogLevel.DEBUG)
    host_str = format_bind_host_literal(host)
    logger.debug(f"Starting Ramalama daemon on {host_str}:{port}...")
    try:
        server = RamalamaServer(host, port, model_store_path, timedelta(seconds=10))
    except OSError as e:
        if host != "::" or e.errno not in (errno.EAFNOSUPPORT, errno.EADDRNOTAVAIL, errno.EINVAL):
            raise
        host = "0.0.0.0"
        logger.debug(f"IPv6 not available, falling back to {host}:{port}...")
        server = RamalamaServer(host, port, model_store_path, timedelta(seconds=10))
    with server:
        with ShutdownHandler(server):
            server_thread = threading.Thread(target=server.serve_forever, daemon=True)
            server_thread.start()
            server_thread.join()


if __name__ == '__main__':
    args = parse_args()
    run(args.host, args.port, args.model_store_path)



from __future__ import annotations

import json
import os
import subprocess
import tempfile
from functools import cached_property
from textwrap import dedent
from typing import Optional, Union

import ramalama.annotations as oci_annotations
from ramalama.common import MNT_DIR, exec_cmd, perror, run_cmd, set_accel_env_vars
from ramalama.engine import BuildEngine, dry_run
from ramalama.oci_tools import OciRef, engine_supports_manifest_attributes
from ramalama.transports.base import Transport
from ramalama.transports.oci import spec as oci_spec
from ramalama.transports.oci.strategies import BaseOCIStrategy
from ramalama.transports.oci.strategy import OCIStrategyFactory

prefix = "oci://"
ociimage_raw = "org.containers.type=ai.image.model.raw"
ociimage_car = "org.containers.type=ai.image.model.car"


class OCI(Transport):
    type = "OCI"

    def format_model(self, model: str) -> str:
        model_name = model.split('/')[-1]
        if ":" not in model_name:
            model = f"{model}:latest"
        return model

    def __init__(self, model: str, model_store_path: str, conman: str, ignore_stderr: bool = False):
        super().__init__(self.format_model(model), model_store_path)

        if not conman:
            raise ValueError("RamaLama OCI Images requires a container engine")
        self.conman = conman
        self.ignore_stderr = ignore_stderr
        self.ref: OciRef = OciRef.from_ref_string(self.model.removeprefix(prefix))

    @cached_property
    def strategy(self) -> BaseOCIStrategy:
        return OCIStrategyFactory(self.conman, model_store=self.model_store).resolve(self.ref)

    @property
    def is_artifact(self) -> bool:
        return self.strategy.kind == 'artifact'

    def login(self, args):
        conman_args = [self.conman, "login"]
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])
        if args.username:
            conman_args.extend([f"--username={args.username}"])
        if args.password:
            conman_args.extend([f"--password={args.password}"])
        if args.passwordstdin:
            conman_args.append("--password-stdin")
        if args.REGISTRY:
            conman_args.append(args.REGISTRY.removeprefix(prefix))
        return exec_cmd(conman_args)

    def logout(self, args):
        conman_args = [self.conman, "logout"]
        conman_args.append(self.model)
        return exec_cmd(conman_args)

    def _target_decompose(self, model):
        model = model.removeprefix(prefix)
        # Remove the prefix and extract target details
        try:
            registry, reference = model.split("/", 1)
        except Exception:
            raise KeyError(
                "You must specify a registry for the model in the form "
                f"'oci://registry.acme.org/ns/repo:tag', got instead: {self.model}"
            )

        reference_dir = reference.replace(":", "/")
        return registry, reference, reference_dir

    def _has_local_snapshot(self) -> bool:
        try:
            _, cached_files, complete = self.model_store.get_cached_files(self.model_tag)
            return complete and bool(cached_files)
        except Exception:
            return False

    def build_image(self, cfile, contextdir, args):
        if args.type == "car":
            parent = args.carimage
            label = ociimage_car
        else:
            parent = "scratch"
            label = ociimage_raw

        footer = dedent(f"""
            FROM {parent}
            LABEL {label}
            COPY --from=build /data/ /
            """).strip()
        full_cfile = cfile + "\n\n" + footer + "\n"
        if args.debug:
            perror(f"Containerfile: \n{full_cfile}")
        engine = BuildEngine(args)
        return engine.build_containerfile(full_cfile, contextdir)

    def _gguf_containerfile(self, model_file_name, args):
        return dedent(f"""
            FROM {args.carimage} AS build
            COPY {model_file_name} /data/models/{model_file_name}
            RUN ln -s {model_file_name} /data/models/model.file
            """).strip()

    def _generate_containerfile(self, source_model, args):
        # Generate the containerfile content
        # Keep this in sync with docs/ramalama-oci.5.md !
        is_car = args.type == "car"
        is_raw = args.type == "raw"
        if args.type == "artifact":
            raise TypeError("artifact handling should not generate containerfiles.")
        if not is_car and not is_raw:
            raise ValueError(f"argument --type: invalid choice: '{args.type}' (choose from artifact,  car, raw)")
        content = [f"FROM {args.carimage} AS build"]
        model_name = source_model.model_name
        ref_file = source_model.model_store.get_ref_file(source_model.model_tag)

        for file in ref_file.files:
            blob_file_path = source_model.model_store.get_blob_file_path(file.hash)
            blob_file_path = os.path.relpath(blob_file_path, source_model.model_store.blobs_directory)
            content.append(f"COPY {blob_file_path} /data/models/{model_name}/{file.name}")
        name = ref_file.model_files[0].name if ref_file.model_files else model_name
        content.append(f"RUN ln -s {model_name}/{name} /data/models/model.file")

        return "\n".join(content)

    def build(self, source_model, args):
        gguf_dir = None
        if getattr(args, "gguf", None):
            perror("Converting to gguf ...")
            gguf_dir = tempfile.TemporaryDirectory(prefix="RamaLama_convert_")
            contextdir = gguf_dir.name
            from ramalama.plugins.loader import get_runtime

            runtime = get_runtime(args.runtime)
            convert_fn = getattr(runtime, "_convert_to_gguf", None)
            if convert_fn is None:
                raise NotImplementedError("Runtime does not support _convert_to_gguf")
            model_file_name = convert_fn(gguf_dir, source_model, args)
            content = self._gguf_containerfile(model_file_name, args)
        else:
            # use blobs directory as context since paths in Containerfile are relative to it
            contextdir = source_model.model_store.blobs_directory
            content = self._generate_containerfile(source_model, args)
        try:
            return self.build_image(content, contextdir, args)
        finally:
            if gguf_dir:
                gguf_dir.cleanup()

    def tag(self, imageid, target, args):
        # Tag imageid with target
        cmd_args = [
            self.conman,
            "tag",
            imageid,
            target,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args)

    def _rm_artifact(self, ignore) -> None:
        rm_cmd = [
            self.conman,
            "artifact",
            "rm",
        ]
        if ignore:
            rm_cmd.append("--ignore")
        rm_cmd.append(self.model)

        run_cmd(
            rm_cmd,
            ignore_all=True,
        )

    def _add_artifact(self, create, name, path, file_name) -> None:
        filepath = oci_spec.normalize_layer_filepath(file_name)
        metadata = oci_spec.FileMetadata.from_path(path, name=file_name).to_json()
        cmd = [
            self.conman,
            "artifact",
            "add",
            "--annotation",
            f"{oci_spec.LAYER_ANNOTATION_FILEPATH}={filepath}",
            "--annotation",
            f"{oci_spec.LAYER_ANNOTATION_FILE_METADATA}={metadata}",
            "--annotation",
            f"{oci_spec.LAYER_ANNOTATION_FILE_MEDIATYPE_UNTESTED}=true",
        ]
        if create:
            cmd.extend(["--replace", "--type", oci_annotations.ArtifactTypeModelManifest])
        else:
            cmd.extend(["--append"])

        cmd.extend([self.model, path])
        run_cmd(
            cmd,
            ignore_stderr=True,
        )

    def _create_artifact(self, source_model, target, args) -> None:
        model_name = source_model.model_name
        ref_file = source_model.model_store.get_ref_file(source_model.model_tag)
        name = ref_file.model_files[0].name if ref_file.model_files else model_name
        create = True
        for file in ref_file.files:
            blob_file_path = source_model.model_store.get_blob_file_path(file.hash)
            self._add_artifact(create, name, blob_file_path, file.name)
            create = False

    def _create_manifest_without_attributes(self, target, imageid, args):
        # Create manifest list for target with imageid
        cmd_args = [
            self.conman,
            "manifest",
            "create",
            target,
            imageid,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args)

    def _create_manifest(self, target, imageid, args):
        self._create_manifest_without_attributes(target, imageid, args)
        if not engine_supports_manifest_attributes(args.engine):
            return

        # Annotate manifest list
        cmd_args = [
            self.conman,
            "manifest",
            "annotate",
            "--annotation",
            f"{oci_annotations.AnnotationModel}=true",
            "--annotation",
            ociimage_car if args.type == "car" else ociimage_raw,
            "--annotation",
            f"{oci_annotations.AnnotationTitle}={args.SOURCE}",
            target,
            imageid,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args, stdout=None)

    def _convert(self, source_model, args):
        set_accel_env_vars()
        perror(
            f"Converting {source_model.model_store.model_name} ({source_model.model_store.model_type}) to "
            f"{self.model_store.model_name} ({self.model_store.model_type}) ..."
        )
        for rm_cmd in [
            [self.conman, "manifest", "rm", self.model],
            [self.conman, "rmi", self.model],
        ]:
            try:
                if args.dryrun:
                    dry_run(rm_cmd)
                else:
                    run_cmd(rm_cmd, ignore_stderr=True, stdout=None)
            except subprocess.CalledProcessError:
                pass
        if args.type == "artifact":
            perror(f"Creating Artifact {self.model} ...")
            self._create_artifact(source_model, self.model, args)
            return

        perror(f"Building {self.model} ...")
        imageid = self.build(source_model, args)
        if args.dryrun:
            imageid = "a1b2c3d4e5f6"
        if self.conman == "docker":
            # docker manifest create doesn't support local image references, so it's not usable here
            self.tag(imageid, self.model, args)
        else:
            self._create_manifest(self.model, f"containers-storage:{imageid}", args)

    def convert(self, source_model, args):
        self._convert(source_model, args)

    def push(self, source_model, args):
        target = self.model
        source = source_model.model
        conman_args = [self.conman, "push"]
        type = "image"
        if args.type == "artifact":
            type = args.type
            conman_args.insert(1, "artifact")

        perror(f"Pushing {type} {self.model} ...")
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        conman_args.extend([target])
        if source != target:
            self._convert(source_model, args)
        try:
            run_cmd(conman_args, ignore_stderr=self.ignore_stderr)
        except subprocess.CalledProcessError as e:
            try:
                if args.type != "artifact":
                    perror(f"Pushing artifact {self.model} ...")
                    conman_args.insert(1, "artifact")
                    run_cmd(conman_args)
            except subprocess.CalledProcessError:
                perror(f"Failed to push OCI {target} : {e}")
                raise e

    def pull(self, args):
        conman_args = []
        if args.quiet:
            conman_args.extend(['--quiet'])
        else:
            perror(f"Downloading {self.model} ...")
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])

        self.strategy.pull(self.ref, cmd_args=conman_args)

    def remove(self, args) -> bool:
        cmd_args = []
        ignore = getattr(args, "ignore", False)
        if ignore:
            if self.strategy.kind == "artifact":
                cmd_args.append("--ignore")
            else:
                cmd_args.append(f"--force={ignore}")
        return self.strategy.remove(self.ref, cmd_args=cmd_args)

    def exists(self) -> bool:
        return self.strategy.exists(self.ref)

    def entrypoint_path(self, mount_dir: Optional[str] = None) -> str:
        return self.strategy.entrypoint_path(self.ref, mount_dir=mount_dir)

    def inspect(
        self,
        show_all: bool = False,
        show_all_metadata: bool = False,
        get_field: str = "",
        as_json: bool = False,
        dryrun: bool = False,
    ) -> None:
        out = super().inspect(
            show_all=show_all, show_all_metadata=show_all_metadata, get_field=get_field, dryrun=dryrun, as_json=as_json
        )

        if get_field or dryrun:
            print(out)
            return

        if as_json:
            out = json.loads(out)

            # docker/podman image inspect returns a list of objects
            inspect = json.loads(self.strategy.inspect(self.ref))
            if self.strategy.kind == "image":
                inspect = inspect[0] if inspect else {}

            print(json.dumps(out | inspect, sort_keys=True, indent=4))
        else:
            print(f"{out}   Type: {self.strategy.kind.capitalize()}")

    def mount_cmd(self, src: Union[str, OciRef, None] = None, dest: Optional[str] = None):
        if isinstance(src, OciRef):
            return self.strategy.mount_arg(src, dest)
        if src is not None:
            return f"--mount=type=volume,src={src},dst={dest or MNT_DIR},readonly"
        return self.strategy.mount_arg(self.ref, dest)



from __future__ import annotations

"""Version of RamaLamaPy."""

__version__ = "0.24.0"


def version():
    return __version__


def print_version(args):
    if args.quiet:
        print(version())
    else:
        print("ramalama version %s" % version())


if __name__ == "__main__":
    print(version())



from __future__ import annotations

import os
from collections.abc import Callable
from typing import Literal, Optional

from ramalama.common import run_cmd
from ramalama.model_store.store import ModelStore
from ramalama.oci_tools import OciRef
from ramalama.transports.oci import spec as oci_spec
from ramalama.transports.oci.oci_artifact import OCIRegistryClient

ReferenceKind = Literal["artifact", "image", "unknown"]


def _format_oci_reference(oci_ref: OciRef) -> str:
    repository = f"{oci_ref.registry}/{oci_ref.repository}"
    if oci_ref.digest:
        return f"{repository}@{oci_ref.digest}"
    tag = oci_ref.tag or "latest"
    return f"{repository}:{tag}"


def fetch_manifest(oci_ref: OciRef) -> Optional[dict]:
    client = OCIRegistryClient(oci_ref.registry, oci_ref.repository, oci_ref.specifier)
    try:
        manifest, _ = client.get_manifest()
        return manifest
    except Exception:
        return None


def manifest_kind(oci_ref: OciRef) -> ReferenceKind:
    manifest = fetch_manifest(oci_ref)
    if not manifest:
        return "unknown"
    if oci_spec.is_cncf_artifact_manifest(manifest):
        return "artifact"
    return "image"


def engine_artifact_exists(engine: str, oci_ref: OciRef, runner: Optional[Callable] = None) -> bool:
    runner = runner or run_cmd
    try:
        runner([engine, "artifact", "inspect", _format_oci_reference(oci_ref)], ignore_stderr=True)
        return True
    except Exception:
        return False


def engine_image_exists(engine: str, oci_ref: OciRef, runner: Optional[Callable] = None) -> bool:
    runner = runner or run_cmd
    try:
        runner([engine, "image", "inspect", _format_oci_reference(oci_ref)], ignore_stderr=True)
        return True
    except Exception:
        return False


def resolve_engine_kind(engine: str, oci_ref: OciRef, runner: Optional[Callable] = None) -> ReferenceKind:
    if not engine:
        return "unknown"
    engine_name = os.path.basename(engine)
    if engine_name == "podman":
        if engine_artifact_exists(engine, oci_ref, runner=runner):
            return "artifact"
        elif engine_image_exists(engine, oci_ref, runner=runner):
            return "image"

    if engine_name == "docker":
        if engine_image_exists(engine, oci_ref, runner=runner):
            return "image"
    return "unknown"


def model_tag_from_reference(oci_ref: OciRef) -> str:
    return oci_ref.specifier


def model_store_has_snapshot(model_store: ModelStore, oci_ref: OciRef) -> bool:
    model_tag = model_tag_from_reference(oci_ref)
    try:
        _, cached_files, complete = model_store.get_cached_files(model_tag)
        return complete and bool(cached_files)
    except Exception:
        return False


class OCITypeResolver:
    def __init__(self, engine: str, model_store: Optional[ModelStore] = None, runner: Optional[Callable] = None):
        self.engine = engine
        self.model_store = model_store
        self.runner = runner or run_cmd

    def resolve(self, reference: OciRef) -> ReferenceKind:
        if self.model_store and model_store_has_snapshot(self.model_store, reference):
            return "artifact"

        engine_kind = resolve_engine_kind(self.engine, reference, runner=self.runner)
        if engine_kind != "unknown":
            return engine_kind

        if "." in reference.registry or ":" in reference.registry or reference.registry == "localhost":
            return manifest_kind(reference)
        return "unknown"



from __future__ import annotations

import json
import os
import urllib.request
from pathlib import Path
from typing import Optional

from ramalama.common import run_cmd
from ramalama.hf_style_repo_base import (
    HFStyleRepoFile,
    HFStyleRepoModel,
    HFStyleRepository,
    fetch_checksum_from_api_base,
)
from ramalama.logger import logger

missing_huggingface = """This operation requires huggingface-cli which is not available.

This tool can be installed via PyPI tools like uv, pip, pip3 or pipx. Example:

pip install -U "huggingface_hub[cli]"

Or via distribution package managers like dnf or apt. Example:

sudo dnf install python3-huggingface-hub
"""


def huggingface_endpoint() -> str:
    """Return Hugging Face base URL from HF_ENDPOINT env var, defaulting to https://huggingface.co"""
    return os.environ.get("HF_ENDPOINT", "").rstrip("/") or HuggingfaceRepository.REGISTRY_URL


def huggingface_token() -> str | None:
    """Return Hugging Face token from HF_TOKEN env var or cached token file, otherwise None"""
    if "HF_TOKEN" in os.environ:
        return os.environ["HF_TOKEN"].strip() or None

    token_path = os.path.expanduser(os.path.join("~", ".cache", "huggingface", "token"))
    if os.path.exists(token_path):
        try:
            with open(token_path) as tokenfile:
                return tokenfile.read().strip()
        except OSError:
            pass

    return None


def extract_huggingface_checksum(data):
    """Extract SHA-256 checksum from Hugging Face API response."""
    # Extract the SHA-256 checksum from the `oid sha256` line
    for line in data.splitlines():
        if line.startswith("oid sha256:"):
            return line.split(":", 1)[1].strip()
    raise ValueError("SHA-256 checksum not found in the API response.")


def fetch_checksum_from_api(organization, file):
    """Fetch the SHA-256 checksum from the model's metadata API for a given file."""
    checksum_api_url = f"{huggingface_endpoint()}/{organization}/raw/main/{file}"
    headers = {}
    token = huggingface_token()
    if token is not None:
        headers['Authorization'] = f"Bearer {token}"

    return fetch_checksum_from_api_base(checksum_api_url, headers, extract_huggingface_checksum)


def fetch_repo_manifest(repo_name: str, tag: str = "latest"):
    # Replicate llama.cpp -hf logic
    # https://github.com/ggml-org/llama.cpp/blob/7f323a589f8684c0eb722e7309074cb5eac0c8b5/common/arg.cpp#L611
    token = huggingface_token()
    repo_manifest_url = f"{huggingface_endpoint()}/v2/{repo_name}/manifests/{tag}"
    logger.debug(f"Fetching repo manifest from {repo_manifest_url}")
    request = urllib.request.Request(
        url=repo_manifest_url,
        headers={
            'User-agent': 'llama-cpp',  # Note: required to return ggufFile field
            'Accept': 'application/json',
        },
    )
    if token is not None:
        request.add_header('Authorization', f"Bearer {token}")

    with urllib.request.urlopen(request) as response:
        repo_manifest = response.read().decode('utf-8')
        return json.loads(repo_manifest)


def fetch_repo_files(repo_name: str, revision: str = "main"):
    """Fetch the list of files in a HuggingFace repository using the Files API with pagination support."""
    token = huggingface_token()
    base_api_url = f"{huggingface_endpoint()}/api/models/{repo_name}/tree/{revision}"

    all_files = []
    next_url: Optional[str] = base_api_url

    # TODO: Handle Diffusers-multifolder layout
    # See https://huggingface.co/docs/diffusers/v0.35.1/using-diffusers/other-formats#diffusers-multifolder
    while next_url:
        logger.debug(f"Fetching repo files from {next_url}")

        request = urllib.request.Request(
            url=next_url,
            headers={
                'Accept': 'application/json',
            },
        )
        if token is not None:
            request.add_header('Authorization', f"Bearer {token}")

        with urllib.request.urlopen(request) as response:
            files_data = response.read().decode('utf-8')
            data = json.loads(files_data)

            # Response should be a list of files or a dict
            if isinstance(data, list):
                all_files.extend(data)
            elif isinstance(data, dict):
                files = data.get('files') or data.get('siblings') or []
                all_files.extend(files)
            else:
                raise ValueError(f"Unexpected response from HuggingFace API: {type(data)}")

            # Check for pagination via Link header
            next_url = None
            link_header = response.headers.get('Link')
            if link_header:
                # Parse Link header: <url>; rel="next"
                # Example: '<https://huggingface.co/api/models/foo/tree/main?cursor=abc>; rel="next"'
                for link in link_header.split(','):
                    parts = link.split(';')
                    has_next_rel = any('rel="next"' in p.strip() or "rel='next'" in p.strip() for p in parts)
                    if has_next_rel:
                        for p in parts:
                            p = p.strip()
                            if p.startswith('<') and p.endswith('>'):
                                next_url = p[1:-1]
                                logger.debug(f"Found next page via Link header: {next_url}")
                                break
                        if next_url:
                            break

    logger.debug(f"Fetched {len(all_files)} total files from repository {repo_name}")
    return all_files


class HuggingfaceCLIFile(HFStyleRepoFile):
    pass


class HuggingfaceRepository(HFStyleRepository):
    REGISTRY_URL = "https://huggingface.co"

    def _fetch_manifest_metadata(self):
        # Repo org/name. Fetch repo manifest to determine model/mmproj file
        self.blob_url = f"{huggingface_endpoint()}/{self.organization}/{self.name}/resolve/main"

        try:
            self.manifest = fetch_repo_manifest(f"{self.organization}/{self.name}", self.tag)
        except (urllib.error.HTTPError, json.JSONDecodeError, KeyError) as e:
            logger.debug(f'fetch_repo_manifest failed: {e}')
            return False
        try:
            # Note that the blobId in the manifest already has a sha256: prefix
            self.model_filename = self.manifest['ggufFile']['rfilename']
            self.model_hash = self.manifest['ggufFile']['blobId']
            self.mmproj_filename = self.manifest.get('mmprojFile', {}).get('rfilename', None)
            self.mmproj_hash = self.manifest.get('mmprojFile', {}).get('blobId', None)
            return True
        except KeyError:
            # No ggufFile in manifest
            return False

    def _collect_file(self, file_list, file_info):
        path = file_info['path']
        oid = file_info.get('oid', '')
        if 'lfs' in file_info and 'oid' in file_info['lfs']:
            oid = file_info['lfs']['oid']
        file_list.append({'filename': path, 'oid': oid})

    def _fetch_safetensors_metadata(self):
        """Fetch metadata for safetensors models from HuggingFace API."""
        repo_name = f"{self.organization}/{self.name}"
        try:
            files = fetch_repo_files(repo_name)
        except (urllib.error.HTTPError, json.JSONDecodeError, KeyError) as e:
            logger.debug(f'fetch_repo_files failed: {e}')
            return False

        # Find all safetensors files, config files and index files
        safetensors_files: list[dict[str, str]] = []
        self.other_files = []
        index_file = None

        try:
            for file_info in files:
                if file_info.get('type') != 'file':
                    continue

                path_str = file_info['path']
                logger.debug(f"Examining file {path_str}")

                path = Path(path_str)
                # Note: case sensitivity follows platform defaults
                if path.match('*.safetensors'):
                    self._collect_file(safetensors_files, file_info)
                elif path.match('*.safetensors.index.json'):
                    index_file = path_str
                elif path_str in {
                    self.FILE_NAME_CONFIG,
                    self.FILE_NAME_GENERATION_CONFIG,
                    self.FILE_NAME_TOKENIZER_CONFIG,
                }:
                    continue
                else:
                    self._collect_file(self.other_files, file_info)

            if not safetensors_files:
                logger.debug('No safetensors files found')
                return False

            # Sort safetensors files by name for consistent ordering
            safetensors_files.sort(key=lambda x: x['filename'])

            # Use the first safetensors file as the main model
            # If there are multiple files, they might be sharded
            self.model_filename = safetensors_files[0]['filename']
            self.model_hash = f"sha256:{safetensors_files[0]['oid']}"

            # Store additional safetensors files for get_file_list
            self.additional_safetensor_files = safetensors_files[1:]

            # Store index file if found
            self.safetensors_index_file = index_file
        except (KeyError, ValueError) as e:
            logger.debug(f'_fetch_safetensors_metadata failed: {e}')
            return False

        return True

    def fetch_metadata(self):
        # Try to fetch GGUF manifest first, then safetensors metadata
        if not self._fetch_manifest_metadata() and not self._fetch_safetensors_metadata():
            raise KeyError("No metadata found")

        token = huggingface_token()
        if token is not None:
            self.headers['Authorization'] = f"Bearer {token}"


class HuggingfaceRepositoryModel(HuggingfaceRepository):
    def fetch_metadata(self):
        # Model url. organization is <org>/<repo>, name is model file path
        self.blob_url = f"{huggingface_endpoint()}/{self.organization}/resolve/main"
        self.model_hash = f"sha256:{fetch_checksum_from_api(self.organization, self.name)}"
        self.model_filename = self.name
        token = huggingface_token()
        if token is not None:
            self.headers['Authorization'] = f"Bearer {token}"


class Huggingface(HFStyleRepoModel):
    REGISTRY_URL = "https://huggingface.co/v2/"
    ACCEPT = "Accept: application/vnd.docker.distribution.manifest.v2+json"

    def __init__(self, model, model_store_path):
        super().__init__(model, model_store_path)

        self.type = "huggingface"

    def get_cli_command(self):
        return "hf"

    def get_login_args(self):
        """HuggingFace CLI uses 'hf auth login' instead of 'hf login'"""
        return ["auth", "login"]

    def get_logout_args(self):
        """HuggingFace CLI uses 'hf auth logout' instead of 'hf logout'"""
        return ["auth", "logout"]

    def get_missing_message(self):
        return missing_huggingface

    def get_registry_url(self):
        return f"{huggingface_endpoint()}/v2/"

    def get_accept_header(self):
        return self.ACCEPT

    def get_repo_type(self):
        return "huggingface"

    def fetch_checksum_from_api(self, organization, file):
        return fetch_checksum_from_api(organization, file)

    def create_repository(self, name, organization, tag):
        if '/' in organization:
            return HuggingfaceRepositoryModel(name, organization, tag)
        else:
            return HuggingfaceRepository(name, organization, tag)

    def get_cli_download_args(self, directory_path, model):
        raise NotImplementedError("huggingface cli download not available")

    def extract_model_identifiers(self):
        model_name, model_tag, model_organization = super().extract_model_identifiers()
        if '/' not in model_organization:
            # if it is a repo then normalize the case insensitive quantization tag
            if model_tag != "latest":
                model_tag = model_tag.upper()
        return model_name, model_tag, model_organization

    def _fetch_snapshot_path(self, cache_dir, namespace, repo):
        cache_path = os.path.join(cache_dir, f'models--{namespace}--{repo}')
        main_ref_path = os.path.join(cache_path, 'refs', 'main')
        if not (os.path.exists(cache_path) and os.path.exists(main_ref_path)):
            return None, None
        with open(main_ref_path, 'r') as file:
            snapshot = file.read().strip()
        snapshot_path = os.path.join(cache_path, 'snapshots', snapshot)
        return snapshot_path, cache_path

    def in_existing_cache(self, args, target_path, sha256_checksum):
        return False

    def push(self, _, args):
        if not getattr(self, "hf_cli_available", False):
            raise NotImplementedError(self.get_missing_message())
        proc = run_cmd(
            [
                "hf",
                "upload",
                "--repo-type",
                "model",
                self.directory,
                self.filename,
                "--cache-dir",
                os.path.join(args.store, "repos", "huggingface", ".cache"),
                "--local-dir",
                os.path.join(args.store, "repos", "huggingface", self.directory),
            ],
        )
        return proc.stdout.decode("utf-8")

    def _collect_cli_files(self, tempdir: str) -> tuple[str, list[HuggingfaceCLIFile]]:
        return "", []



"""CLI registration for the ``ramalama rag`` subcommand."""

from ramalama.cli import (
    OverrideDefaultAction,
    add_engine_args_argument,
    default_image,
    default_rag_image,
    local_images,
    suppressCompleter,
)
from ramalama.config import ActiveConfig
from ramalama.plugins.runtimes.inference.llama_cpp import AddPathOrUrl
from ramalama.plugins.runtimes.inference.rag.handler import DEFAULT_CAPTION_MODEL


def register_rag_subcommand(plugin, subparsers):
    """Register the ``rag`` subcommand on the given subparsers action."""
    rt_config = plugin.get_runtime_config(ActiveConfig())
    parser = subparsers.add_parser(
        "rag",
        help="convert documents to a RAG vector database and package as a container image",
    )
    parser.add_argument(
        "DOCUMENTS",
        nargs="+",
        help="files, directories, or URLs of documents to process",
        action=AddPathOrUrl,
    )
    parser.add_argument(
        "DESTINATION",
        help="name for the output container image containing the Qdrant vector database",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "--docling-model",
        dest="docling_model",
        default="hf://ibm-granite/granite-docling-258M-GGUF",
        help="Granite Docling GGUF model for document conversion",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "--image",
        default=default_image(),
        help="OCI container image to use for the llama.cpp inference server",
        action=OverrideDefaultAction,
        completer=local_images,
    )
    parser.add_argument(
        "-c",
        "--ctx-size",
        dest="ctx_size",
        type=int,
        default=8192,
        help="context size for the VLM server (default: 8192)",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "--embed-ctx-size",
        dest="embed_ctx_size",
        type=int,
        default=0,
        help="context size for the embedding server (default: 0, auto-detected by llama.cpp)",
        completer=suppressCompleter,
    )
    add_engine_args_argument(parser)
    parser.add_argument(
        "--chunk-size",
        dest="chunk_size",
        type=int,
        default=400,
        help="max tokens per chunk for embedding (default: 400)",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "--ngl",
        dest="ngl",
        default=None,
        help="number of layers to store in VRAM: a number, 'auto', or 'all' (default: auto)",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "--rag-image",
        default=default_rag_image(),
        help="OCI container image for the RAG processing container",
        action=OverrideDefaultAction,
        completer=local_images,
    )
    parser.add_argument(
        "--caption-images",
        dest="caption_images",
        nargs="?",
        const=DEFAULT_CAPTION_MODEL,
        default=None,
        metavar="MODEL",
        help=f"enable image captioning via a VLM (default model: {DEFAULT_CAPTION_MODEL})",
        completer=suppressCompleter,
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=rt_config.threads,
        help=f"number of CPU threads to use (default: {rt_config.threads})",
        completer=suppressCompleter,
    )
    parser.set_defaults(func=lambda args: _rag_dispatch(plugin, args))


def _rag_dispatch(plugin, args):
    """Lazy import and dispatch to the RAG handler."""
    from ramalama.plugins.runtimes.inference.rag.handler import rag_handler

    rag_handler(plugin, args)



from __future__ import annotations

import os
import platform
from typing import Optional, Tuple

from ramalama.common import MNT_DIR, RAG_DIR, ContainerEntryPoint, check_nvidia, get_accel_env_vars, get_gpu_devices
from ramalama.file import PlainFile
from ramalama.path_utils import normalize_host_path_for_container
from ramalama.version import version


class Kube:
    def __init__(
        self,
        model_name: str,
        model_paths: Tuple[str, str],
        chat_template_paths: Optional[Tuple[str, str]],
        mmproj_paths: Optional[Tuple[str, str]],
        args,
        exec_args,
        draft_model_paths: Optional[Tuple[str, str]],
        artifact: bool,
    ):
        self.src_chat_template_path, self.dest_chat_template_path = (
            chat_template_paths if chat_template_paths is not None else ("", "")
        )
        self.src_mmproj_path, self.dest_mmproj_path = mmproj_paths if mmproj_paths is not None else ("", "")
        self.model_paths = {'model': model_paths}
        if draft_model_paths is not None:
            self.model_paths['model-draft'] = draft_model_paths
        self.ai_image = model_name
        if getattr(args, "name", None):
            self.name = args.name
        else:
            self.name = "ramalama"

        self.args = args
        self.exec_args = exec_args
        self.image = args.image
        self.artifact = artifact

    def _gen_volumes(self) -> Tuple[str, str]:
        mounts = """\
        volumeMounts:"""

        volumes = """\
      volumes:"""
        for volume_name, (src_model_path, dest_model_path) in self.model_paths.items():
            src_model_path = src_model_path.removeprefix("oci://")
            if os.path.exists(src_model_path):
                m, v = self._gen_path_volume(volume_name, src_model_path, dest_model_path)
                mounts += m
                volumes += v
            else:
                subPath = ""
                if not self.artifact:
                    subPath = """
          subPath: /models"""
                mounts += f"""
        - mountPath: {MNT_DIR}{subPath}
          name: {volume_name}"""
                volumes += self._gen_oci_volume(volume_name, src_model_path)

        if getattr(self.args, 'rag', None):
            m, v = self._gen_rag_volume()
            mounts += m
            volumes += v

        if self.src_chat_template_path and os.path.exists(self.src_chat_template_path):
            m, v = self._gen_chat_template_volume()
            mounts += m
            volumes += v

        if self.src_mmproj_path and os.path.exists(self.src_mmproj_path):
            m, v = self._gen_mmproj_volume()
            mounts += m
            volumes += v

        m, v = self._gen_devices()
        mounts += m
        volumes += v
        return mounts, volumes

    def _gen_devices(self) -> Tuple[str, str]:
        mounts = ""
        volumes = ""
        for name, path in get_gpu_devices().items():
            mounts += f"""
        - mountPath: {path}
          name: {name}"""
            volumes += f"""
      - hostPath:
          path: {path}
        name: {name}"""
        return mounts, volumes

    def _gen_path_volume(self, volume_name, src_model_path, dest_model_path) -> Tuple[str, str]:
        host_model_path = normalize_host_path_for_container(src_model_path)
        if platform.system() == "Windows":
            #  Workaround https://github.com/containers/podman/issues/16704
            host_model_path = '/mnt' + host_model_path
        mount = f"""
        - mountPath: {dest_model_path}
          name: {volume_name}"""
        volume = f"""
      - hostPath:
          path: {host_model_path}
        name: {volume_name}"""
        return mount, volume

    def _gen_oci_volume(self, volume_name, src_model_path) -> str:
        return f"""
      - image:
          reference: {src_model_path}
          pullPolicy: IfNotPresent
        name: {volume_name}"""

    def _gen_rag_volume(self) -> Tuple[str, str]:
        mounts = f"""
        - mountPath: {RAG_DIR}
          name: rag"""

        volumes = f"""
      - image:
          reference: {self.args.rag}
          pullPolicy: IfNotPresent
        name: rag"""

        return mounts, volumes

    def _gen_chat_template_volume(self) -> Tuple[str, str]:
        host_chat_template_path = normalize_host_path_for_container(self.src_chat_template_path)
        if platform.system() == "Windows":
            #  Workaround https://github.com/containers/podman/issues/16704
            host_chat_template_path = '/mnt' + host_chat_template_path
        mount = f"""
        - mountPath: {self.dest_chat_template_path}
          name: chat-template"""
        volume = f"""
      - hostPath:
          path: {host_chat_template_path}
        name: chat-template"""
        return mount, volume

    def _gen_mmproj_volume(self) -> Tuple[str, str]:
        host_mmproj_path = normalize_host_path_for_container(self.src_mmproj_path)
        if platform.system() == "Windows":
            #  Workaround https://github.com/containers/podman/issues/16704
            host_mmproj_path = '/mnt' + host_mmproj_path
        mount = f"""
        - mountPath: {self.dest_mmproj_path}
          name: mmproj"""
        volume = f"""
      - hostPath:
          path: {host_mmproj_path}
        name: mmproj"""
        return mount, volume

    def __gen_ports(self) -> str:
        if not hasattr(self.args, "port"):
            return ""

        p = self.args.port.split(":", 2)
        ports = f"""\
        ports:
        - containerPort: {p[0]}"""
        if len(p) > 1:
            ports += f"""
          hostPort: {p[1]}"""

        return ports

    def __gen_env_vars(self) -> str:
        env_vars = get_accel_env_vars()

        if hasattr(self.args, "env"):
            extra_env = dict(i.split("=", 1) for i in self.args.env)
            env_vars |= extra_env

        if not env_vars:
            return ""

        env_spec = """\
        env:"""

        for k, v in sorted(env_vars.items()):
            env_spec += f"""
        - name: {k}
          value: \"{v}\""""

        return env_spec

    def __gen_security_context(self) -> str:
        return """\
        securityContext:
          allowPrivilegeEscalation: false
          capabilities:
            drop:
            - CAP_CHOWN
            - CAP_FOWNER
            - CAP_FSETID
            - CAP_KILL
            - CAP_NET_BIND_SERVICE
            - CAP_SETFCAP
            - CAP_SETGID
            - CAP_SETPCAP
            - CAP_SETUID
            - CAP_SYS_CHROOT
            add:
            - CAP_DAC_OVERRIDE
          seLinuxOptions:
            type: spc_t"""

    def __gen_resources(self) -> str:
        if check_nvidia() == "cuda":
            return """
        resources:
          limits:
             'nvidia.com/gpu=all': 1"""

        devices = get_gpu_devices()
        if devices:
            limits = "".join(f"\n             'podman.io/device={path}': 1" for path in devices.values())
            return f"""
        resources:
          limits:{limits}"""

        return ""

    def __gen_container(self, container_args) -> str:
        content = f"""\
      - name: {container_args["name"]}
        image: {container_args["image"]}
"""
        if "command" in container_args:
            content += f"""\
        command: ["{container_args["command"]}"]
"""
        content += f"""\
        args: {container_args["args"]}
{container_args["env_string"]}"""
        content += f"""
{self.__gen_security_context()}
{container_args["port_string"]}"""
        if "mounts_string" in container_args:
            content += f"""
{container_args["mounts_string"]}"""
        if "resources_string" in container_args:
            content += f"""\
{container_args["resources_string"]}"""
        return content

    def generate_content(self, name=None, labels="", container=None) -> str:
        env_string = self.__gen_env_vars()
        port_string = self.__gen_ports()
        mounts_string, volume_string = self._gen_volumes()
        _version = version()
        model_container = {
            "name": self.name if name is None else name,
            "image": self.image,
            "args": self.exec_args[1:],
            "env_string": env_string,
            "port_string": port_string,
            "mounts_string": mounts_string,
            "resources_string": self.__gen_resources(),
        }
        if not isinstance(self.exec_args[0], ContainerEntryPoint):
            model_container["command"] = self.exec_args[0]

        content = f"""\
# Save the output of this file and use kubectl create -f to import
# it into Kubernetes.
#
# Created with ramalama-{_version}
apiVersion: apps/v1
kind: Deployment
metadata:
  name: {self.name}
  labels:
    app: {self.name}
spec:
  replicas: 1
  selector:
    matchLabels:
      app: {self.name}
  template:
    metadata:
      labels:
        app: {self.name}{labels}
    spec:
      containers:
{self.__gen_container(model_container)}
"""
        if container is not None:
            content += f"""{self.__gen_container(container)}
"""
        content += f"""{volume_string}
"""

        return content

    def generate(self) -> PlainFile:
        return genfile(self.name, self.generate_content())


def genfile(name, content) -> PlainFile:
    file_name = f"{name}.yaml"
    print(f"Generating Kubernetes YAML file: {file_name}")

    file = PlainFile(file_name)
    file.content = content
    return file



from __future__ import annotations

from ramalama.file_loaders.file_types.base import BaseFileLoader


class TXTFileLoader(BaseFileLoader):
    @staticmethod
    def file_extensions() -> set[str]:
        return {
            ".txt",
            ".sh",
            ".md",
            ".yaml",
            ".yml",
            ".json",
            ".csv",
            ".toml",
        }

    @staticmethod
    def load(file: str) -> str:
        """
        Load the content of the text file.
        """

        # TODO: Support for non-default encodings?
        with open(file, "r") as f:
            return f.read()



from __future__ import annotations


class MissingArgumentError(Exception):
    """Exception raised when a required argument is missing."""

    def __init__(self, field: str):
        super().__init__(f"Missing required argument: {field}")



from __future__ import annotations

from abc import ABC, abstractmethod


class BaseFileLoader(ABC):
    """
    Base class for file upload handlers.
    This class should be extended by specific file type handlers.
    """

    @staticmethod
    @abstractmethod
    def load(file: str) -> str:
        """
        Load the content of the file.
        This method should be implemented by subclasses to handle specific file types.
        """
        pass

    @staticmethod
    @abstractmethod
    def file_extensions() -> set[str]:
        """
        Get the file extension supported by this file type handler.
        """
        pass



from __future__ import annotations

import json
import subprocess
from dataclasses import dataclass
from datetime import datetime
from itertools import chain
from typing import Optional, TypedDict

import ramalama.annotations as oci_annotations
from ramalama.arg_types import EngineArgType
from ramalama.common import SemVer, engine_version, run_cmd
from ramalama.logger import logger

ocilabeltype = "org.containers.type"


def convert_from_human_readable_size(input) -> int:
    """
    Convert a human-readable size string (e.g. '1.5MB', '10KB', '42') to a size in bytes.
    The return type is always an integer number of bytes. Fractional values are
    rounded to the nearest integer instead of being truncated, to avoid silently
    losing data (e.g. '1.5MB' is treated as 1_572_864 bytes).
    """
    sizes = [("KB", 1024), ("MB", 1024**2), ("GB", 1024**3), ("TB", 1024**4), ("B", 1)]
    value = str(input).strip()
    lower_value = value.lower()
    for unit, size in sizes:
        if lower_value.endswith(unit.lower()):
            number_part = value[: -len(unit)].strip()
            return int(round(float(number_part) * size))

    return int(round(float(value)))


def parse_datetime(date_str: str) -> Optional[datetime]:
    try:
        parsed_date = datetime.fromisoformat(date_str.replace(" UTC", "").replace("+0000", "+00:00").replace(" ", "T"))
    except ValueError:
        parsed_date = None

    return parsed_date


class ListModelResponse(TypedDict):
    modified: Optional[datetime]
    name: str
    size: int


def list_artifacts(args: EngineArgType):
    if args.engine is None:
        raise ValueError("Cannot list artifacts without a provided engine like podman or docker.")

    if args.engine == "docker":
        return []

    conman_args = [
        args.engine,
        "artifact",
        "ls",
        "--format",
        (
            '{"name":"oci://{{ .Repository }}:{{ .Tag }}",\
            "created":"{{ .CreatedAt }}", \
            "size":"{{ .Size }}", \
            "ID":"{{ .Digest }}"},'
        ),
    ]
    try:
        if (output := run_cmd(conman_args, ignore_stderr=True).stdout.decode("utf-8").strip()) == "":
            return []
    except subprocess.CalledProcessError as e:
        logger.debug(e)
        return []
    if output == "":
        return []

    try:
        artifacts = json.loads(f"[{output[:-1]}]")
    except json.JSONDecodeError:
        return []

    models = []
    for artifact in artifacts:
        conman_args = [
            args.engine,
            "artifact",
            "inspect",
            artifact["ID"],
        ]
        try:
            output = run_cmd(conman_args, ignore_stderr=True).stdout.decode("utf-8").strip()
        except Exception:
            continue

        if output == "":
            continue
        try:
            inspect = json.loads(output)
        except json.JSONDecodeError:
            continue
        if "Manifest" not in inspect:
            continue
        if "artifactType" not in inspect["Manifest"]:
            continue
        if inspect["Manifest"]['artifactType'] != oci_annotations.ArtifactTypeModelManifest:
            continue
        models.append(
            {
                "name": artifact["name"],
                "modified": parse_datetime(artifact["created"]),
                "size": convert_from_human_readable_size(artifact["size"]),
            }
        )
    return models


def engine_supports_manifest_attributes(engine) -> bool:
    if not engine or engine == "" or engine == "docker":
        return False
    if engine == "podman":
        try:
            if engine_version(engine) < SemVer(5, 0, 0):
                return False
        except Exception:
            return False
    return True


def list_manifests(args: EngineArgType) -> list[ListModelResponse]:
    if args.engine is None:
        raise ValueError("Cannot list manifests without a provided engine like podman or docker.")

    if args.engine == "docker":
        return []

    conman_args = [
        args.engine,
        "images",
        "--filter",
        "manifest=true",
        "--format",
        (
            '{"name":"oci://{{ .Repository }}:{{ .Tag }}","modified":"{{ .CreatedAt }}",'
            '"size":{{ .VirtualSize }}, "ID":"{{ .ID }}"},'
        ),
    ]
    output = run_cmd(conman_args).stdout.decode("utf-8").strip()
    if output == "":
        return []

    manifests = json.loads(f"[{output[:-1]}]")
    if not engine_supports_manifest_attributes(args.engine):
        return [
            {
                "name": manifest["name"],
                "modified": parse_datetime(manifest["modified"]),
                "size": int(manifest["size"]),
            }
            for manifest in manifests
        ]

    models: list[ListModelResponse] = []
    for manifest in manifests:
        conman_args = [
            args.engine,
            "manifest",
            "inspect",
            manifest["ID"],
        ]
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()

        if output == "":
            continue
        inspect = json.loads(output)
        if 'manifests' not in inspect:
            continue
        if not inspect['manifests']:
            continue
        img = inspect['manifests'][0]
        if 'annotations' not in img:
            continue
        if oci_annotations.AnnotationModel in img['annotations']:
            models.append(
                {
                    "name": manifest["name"],
                    "modified": parse_datetime(manifest["modified"]),
                    "size": manifest["size"],
                }
            )
    return models


def list_images(args: EngineArgType) -> list[ListModelResponse]:
    # if engine is docker, size will be retrieved from the inspect command later
    # if engine is podman use "size":{{ .VirtualSize }}
    conman = args.engine
    if conman is None:
        return []

    formatLine = '{"name":"oci://{{ .Repository }}:{{ .Tag }}","modified":"{{ .CreatedAt }}"'
    if conman == "podman":
        formatLine += ',"size":{{ .VirtualSize }}},'
    else:
        formatLine += ',"id":"{{ .ID }}"},'

    conman_args = [
        conman,
        "images",
        "--filter",
        f"label={ocilabeltype}",
        "--format",
        formatLine,
    ]
    if conman == "docker":
        conman_args.insert(2, "--no-trunc")

    output = run_cmd(conman_args, env={"TZ": "UTC"}).stdout.decode("utf-8").strip()
    if output == "":
        return []

    raw = [model for model in json.loads(f"[{output[:-1]}]") if model["name"] != "oci://<none>:<none>"]

    if conman == 'docker':
        ids = [m["id"] for m in raw]
        inspect_args = [conman, "image", "inspect", *ids, "--format", "{{.Id}} {{.Size}}"]
        inspect_out = run_cmd(inspect_args).stdout.decode("utf-8").strip()

        size_by_id: dict[str, str] = {}
        for line in inspect_out.splitlines():
            if not line:
                continue
            image_id, size = line.split(maxsplit=1)
            size_by_id[image_id] = size

        return [
            {
                "name": m["name"],
                "modified": parse_datetime(m["modified"]),
                "size": int(size_by_id[m["id"]]),
            }
            for m in raw
        ]

    return [
        {
            "name": m["name"],
            "modified": parse_datetime(m["modified"]),
            "size": int(m["size"]),
        }
        for m in raw
    ]


def list_models(args: EngineArgType) -> list[ListModelResponse]:
    if args.engine is None:
        return []

    model_gen = chain(list_images(args), list_manifests(args), list_artifacts(args))

    seen: set[str] = set()
    models: list[ListModelResponse] = []
    for m in model_gen:
        if (name := m["name"]) in seen:
            continue
        seen.add(name)
        models.append(m)
    return models


@dataclass(frozen=True)
class OciRef:
    registry: str
    repository: str
    specifier: str  # Either the digest or the tag
    tag: Optional[str] = None
    digest: Optional[str] = None

    def __str__(self) -> str:
        if self.digest:
            return f"{self.registry}/{self.repository}@{self.digest}"
        return f"{self.registry}/{self.repository}:{self.tag or self.specifier}"

    @staticmethod
    def from_ref_string(ref: str) -> "OciRef":
        return split_oci_reference(ref)


def split_oci_reference(ref: str, default_registry: str = "docker.io") -> OciRef:
    ref = ref.strip()

    name, digest = ref.split("@", 1) if "@" in ref else (ref, None)

    slash = name.rfind("/")
    colon = name.rfind(":")
    if colon > slash:
        name, tag = name[:colon], name[colon + 1 :]
    else:
        tag = None

    parts = name.split("/", 1)
    if len(parts) == 1:
        registry = default_registry
        repository = parts[0]
    else:
        first, rest = parts[0], parts[1]
        if first == "localhost" or "." in first or ":" in first:
            registry = first
            repository = rest
        else:
            registry = default_registry
            repository = name  # keep full path

    specifier = digest or tag
    if specifier is None:
        tag = "latest"
        specifier = tag

    return OciRef(registry=registry, repository=repository, tag=tag, digest=digest, specifier=specifier)



from __future__ import annotations

import copy
import os

import ramalama.kube as kube
import ramalama.quadlet as quadlet
from ramalama.common import exec_cmd, genname, get_accel_env_vars
from ramalama.compat import NamedTemporaryFile
from ramalama.compose import Compose
from ramalama.config import ActiveConfig, Config
from ramalama.engine import add_labels
from ramalama.kube import Kube
from ramalama.plugins.loader import assemble_command
from ramalama.transports.base import compute_serving_port
from ramalama.transports.transport_factory import New


def stack_image(config: Config) -> str:
    return config.stack_image or config.default_stack_image


class Stack:
    """Stack class"""

    type = "Stack"

    def __init__(self, args):
        self.args = args
        self.name = getattr(args, "name", None) or genname()
        self.host = "::"
        self.model = New(args.MODEL, args)
        self.model_type = self.model.type
        self.model_port = "8080"
        self.stack_image = stack_image(ActiveConfig())
        self.labels = ""
        self.draft_model = New(args.model_draft, args) if args.model_draft is not None else None

    def add_label(self, label):
        cleanlabel = label.replace("=", ": ", 1)
        self.labels = f"{self.labels}\n        {cleanlabel}"

    def _get_env_vars(self) -> dict:
        env_vars = {}
        if hasattr(self.args, "env"):
            for e in self.args.env:
                env = e.split("=", 1)
                env_vars[env[0]] = env[1]
        return env_vars

    def _gen_compose_env(self, env_vars) -> str:
        compose_env = ""
        for k, v in env_vars.items():
            compose_env += f"""\n      - {k}={v}"""
        return compose_env

    def _gen_kube_env(self, env_vars) -> str:
        kube_env = ""
        for k, v in env_vars.items():
            kube_env += f"""\n        - name: {k}
          value: {v}"""
        return kube_env

    def _gen_server_env(self) -> str:
        accel_env_vars = get_accel_env_vars()
        if "MUSA_VISIBLE_DEVICES" in accel_env_vars:
            accel_env_vars["MTHREADS_VISIBLE_DEVICES"] = "all"
        return self._gen_kube_env(accel_env_vars)

    def generate(self) -> str:
        add_labels(self.args, self.add_label)
        env_vars = self._get_env_vars()
        common_env = self._gen_kube_env(env_vars)
        llama_stack_container = {
            "name": "llama-stack",
            "image": f"{self.stack_image}",
            "args": [],
            "env_string": f"""\
        env:{common_env}
        - name: RAMALAMA_URL
          value: http://127.0.0.1:{self.model_port}
        - name: RAMALAMA_RUNTIME
          value: \"{self.args.runtime}\"
        - name: INFERENCE_MODEL
          value: \"{self.model.model_alias}\"""",
            "port_string": f"""\
        ports:
        - containerPort: 8321
          hostPort: {self.args.port}""",
        }
        kube = Kube(*(self._delegate_args() + tuple([False])))
        return kube.generate_content("model-server", self.labels, llama_stack_container)

    def _delegate_args(self):
        model_src_path = self.model._get_entry_model_path(False, False, False)
        chat_template_src_path = self.model._get_chat_template_path(False, False, False)
        mmproj_src_path = self.model._get_mmproj_path(False, False, False)
        model_dest_path = self.model._get_entry_model_path(True, True, False)
        chat_template_dest_path = self.model._get_chat_template_path(True, True, False)
        mmproj_dest_path = self.model._get_mmproj_path(True, True, False)
        args2 = copy.copy(self.args)
        args2.port = self.model_port
        exec_args = assemble_command(args2)
        draft_model_paths = None
        if self.draft_model is not None:
            draft_model_paths = (
                self.draft_model._get_entry_model_path(False, False, False),
                self.draft_model._get_entry_model_path(True, True, False),
            )

        return (
            self.model.model_name,
            (model_src_path, model_dest_path),
            (chat_template_src_path, chat_template_dest_path),
            (mmproj_src_path, mmproj_dest_path),
            args2,
            exec_args,
            draft_model_paths,
        )

    def serve(self):
        stack_port = compute_serving_port(self.args, quiet=self.args.generate)
        self.args.port = stack_port
        if self.args.dryrun:
            yaml = self.generate()
            print(yaml)
            return

        if self.args.generate:
            if self.args.generate.gen_type == "kube":
                yaml = self.generate()
                kube.genfile(self.name, yaml).write(self.args.generate.output_dir)
                return

            if self.args.generate.gen_type == "quadlet/kube" or self.args.generate.gen_type == "quadlet":
                yaml = self.generate()
                kube.genfile(self.name, yaml).write(self.args.generate.output_dir)
                k = quadlet.kube(
                    self.name, f"RamaLama {self.model.model_alias} Kubernetes YAML - llama Stack AI Model Service"
                )
                openai = f"http://localhost:{self.args.port}"
                k.add("comment", f"# RamaLama service for {self.model.model_alias}")
                k.add("comment", "# Serving RESTAPIs:")
                k.add("comment", f"#    Llama Stack: {openai}")
                k.add("comment", f"#    OpenAI:      {openai}/v1/openai\n")
                k.write(self.args.generate.output_dir)
                return

            if self.args.generate.gen_type == "compose":
                compose = Compose(*self._delegate_args())
                file = compose.generate()
                compose_env = self._gen_compose_env(self._get_env_vars())
                file.content += f"""
  {self.model.model_name}-stack:
    image: {self.stack_image}
    container_name: {self.name}-stack
    ports:
      - "{stack_port}:8123"
    environment:{compose_env}
      - RAMALAMA_URL=http://{self.name}:{self.model_port}
      - RAMALAMA_RUNTIME={self.args.runtime}
      - INFERENCE_MODEL={self.model.model_alias}
    depends_on:
      - {self.model.model_name}
    restart: unless-stopped"""
                file.write(self.args.generate.output_dir)
                return

        if not os.path.basename(self.args.engine).startswith("podman"):
            raise ValueError("llama-stack requires use of the Podman container engine")
        with NamedTemporaryFile(
            mode='w', prefix='RamaLama_', delete=not self.args.debug, delete_on_close=False
        ) as yaml_file:
            yaml = self.generate()
            yaml_file.write(yaml)
            yaml_file.close()

            exec_args = [
                self.args.engine,
                "kube",
                "play",
                "--replace",
            ]
            if not self.args.detach:
                exec_args.append("--wait")

            extra = getattr(self.args, "engine_args", None) or []
            if extra:
                exec_args.extend(extra)

            exec_args.append(yaml_file.name)
            exec_cmd(exec_args)

    def stop(self):
        if not os.path.basename(self.args.engine).startswith("podman"):
            raise ValueError("llama-stack requires use of the Podman container engine")
        with NamedTemporaryFile(
            mode='w', prefix='RamaLama_', delete=not self.args.debug, delete_on_close=False
        ) as yaml_file:
            yaml_file.write(self.generate())
            yaml_file.close()

            exec_args = [
                self.args.engine,
                "kube",
                "down",
            ]
            extra = getattr(self.args, "engine_args", None) or []
            if extra:
                exec_args.extend(extra)
            exec_args.append(yaml_file.name)
            exec_cmd(exec_args)



from __future__ import annotations

import base64
import re
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Any, Literal, Optional, Protocol, Union

from ramalama.console import should_colorize

# Strip ANSI escape sequences and control chars to prevent terminal injection (e.g. from LLM output)
_ANSI_ESCAPE_RE = re.compile(r"\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~]|\][^\x1b]*(?:\x1b\\|\x07))")
_CONTROL_CHARS_RE = re.compile(r"[\x00-\x08\x0b\x0c\x0e-\x1f\x7f]")


def sanitize_for_terminal(text: str) -> str:
    """Remove ANSI escape sequences and control characters from untrusted output before printing."""
    if not text:
        return text
    s = _ANSI_ESCAPE_RE.sub("", text)
    return _CONTROL_CHARS_RE.sub("", s)


RoleType = Literal["system", "user", "assistant", "tool"]


@dataclass
class ImageURLPart:
    url: str
    detail: Optional[str] = None
    type: Literal["image_url"] = "image_url"


@dataclass
class ImageBytesPart:
    data: bytes
    mime_type: str = "application/octet-stream"
    type: Literal["image_bytes"] = "image_bytes"


@dataclass
class ToolCall:
    id: str
    name: str
    arguments: dict[str, Any]


AttachmentPart = Union[ImageURLPart, ImageBytesPart]


@dataclass
class SystemMessage:
    role: Literal["system"] = "system"
    text: str = ""
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class UserMessage:
    role: Literal["user"] = "user"
    text: Optional[str] = None
    metadata: dict[str, Any] = field(default_factory=dict)
    attachments: list[AttachmentPart] = field(default_factory=list)


@dataclass
class AssistantMessage:
    role: Literal["assistant"] = "assistant"
    text: Optional[str] = None
    metadata: dict[str, Any] = field(default_factory=dict)
    tool_calls: list[ToolCall] = field(default_factory=list)
    attachments: list[AttachmentPart] = field(default_factory=list)


@dataclass
class ToolMessage:
    text: str
    role: Literal["tool"] = "tool"
    metadata: dict[str, Any] = field(default_factory=dict)
    tool_call_id: Optional[str] = None


ChatMessageType = Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage]


class StreamParser(Protocol):
    def parse_stream_chunk(self, chunk: bytes) -> Iterable[Any]: ...


def stream_response(chunks: Iterable[bytes], color: str, provider: StreamParser) -> str:
    color_default = ""
    color_yellow = ""
    if (color == "auto" and should_colorize()) or color == "always":
        color_default = "\033[0m"
        color_yellow = "\033[33m"

    print("\r", end="")
    assistant_response = ""
    for chunk in chunks:
        events = provider.parse_stream_chunk(chunk)
        for event in events:
            text = getattr(event, "text", None)
            if not text:
                continue
            safe_text = sanitize_for_terminal(text)
            print(f"{color_yellow}{safe_text}{color_default}", end="", flush=True)
            assistant_response += text

    print("")
    return assistant_response


def serialize_part(part: AttachmentPart) -> dict[str, Any]:
    if isinstance(part, ImageURLPart):
        payload: dict[str, Any] = {"url": part.url}
        if part.detail:
            payload["detail"] = part.detail
        return {"type": "image_url", "image_url": payload}
    if isinstance(part, ImageBytesPart):
        return {
            "type": "image_bytes",
            "image_bytes": {"data": base64.b64encode(part.data).decode("ascii"), "mime_type": part.mime_type},
        }

    raise TypeError(f"Unsupported message part: {part!r}")


__all__ = [
    "ToolCall",
    "ImageURLPart",
    "ImageBytesPart",
    "sanitize_for_terminal",
    "stream_response",
    "serialize_part",
]



from __future__ import annotations

# ramalama/generate/compose.py
import os
import shlex
from typing import Optional

from ramalama.common import RAG_DIR, get_accel_env_vars, get_gpu_devices
from ramalama.file import PlainFile
from ramalama.version import version


class Compose:
    def __init__(
        self,
        model_name: str,
        model_paths: tuple[str, str],
        chat_template_paths: Optional[tuple[str, str]],
        mmproj_paths: Optional[tuple[str, str]],
        args,
        exec_args,
        draft_model_paths: Optional[tuple[str, str]],
    ):
        self.src_model_path, self.dest_model_path = model_paths
        self.src_draft_model_path, self.dest_draft_model_path = (
            draft_model_paths if draft_model_paths is not None else ("", "")
        )
        self.src_chat_template_path, self.dest_chat_template_path = (
            chat_template_paths if chat_template_paths is not None else ("", "")
        )
        self.src_mmproj_path, self.dest_mmproj_path = mmproj_paths if mmproj_paths is not None else ("", "")
        self.src_model_path = self.src_model_path.removeprefix("oci://")
        if self.src_draft_model_path:
            self.src_draft_model_path = self.src_draft_model_path.removeprefix("oci://")

        self.model_name = model_name
        custom_name = getattr(args, "name", None)
        self.name = custom_name if custom_name else f"ramalama-{model_name}"
        self.args = args
        self.exec_args = exec_args
        self.image = args.image

    def _gen_volumes(self) -> str:
        volumes = "    volumes:"

        # Model Volume
        volumes += self._gen_model_volume(self.src_model_path, self.dest_model_path)
        if self.src_draft_model_path and self.dest_draft_model_path:
            volumes += self._gen_model_volume(self.src_draft_model_path, self.dest_draft_model_path)

        # RAG Volume
        if getattr(self.args, "rag", None):
            volumes += self._gen_rag_volume()

        # Chat Template Volume
        if self.src_chat_template_path and os.path.exists(self.src_chat_template_path):
            volumes += self._gen_chat_template_volume()

        # MMProj Volume
        if self.src_mmproj_path and os.path.exists(self.src_mmproj_path):
            volumes += self._gen_mmproj_volume()

        return volumes

    def _gen_model_volume(self, src_model_path: str, dest_model_path: str) -> str:
        return f'\n      - "{src_model_path}:{dest_model_path}:ro"'

    def _gen_rag_volume(self) -> str:
        rag_source = self.args.rag
        volume_str = ""

        if rag_source.startswith("oci:") or rag_source.startswith("oci://"):
            if rag_source.startswith("oci://"):
                oci_image = rag_source.removeprefix("oci://")
            else:
                oci_image = rag_source.removeprefix("oci:")
            # This is the standard long-form syntax for image volumes, now supported by Docker.
            volume_str = f"""
      - type: image
        source: {oci_image}
        target: {RAG_DIR}
        image:
          readonly: true"""

        elif os.path.exists(rag_source):
            # Standard host path mount
            volume_str = f'\n      - "{rag_source}:{RAG_DIR}:ro"'

        return volume_str

    def _gen_chat_template_volume(self) -> str:
        return f'\n      - "{self.src_chat_template_path}:{self.dest_chat_template_path}:ro"'

    def _gen_mmproj_volume(self) -> str:
        return f'\n      - "{self.src_mmproj_path}:{self.dest_mmproj_path}:ro"'

    def _gen_devices(self) -> str:
        devices = get_gpu_devices()

        if not devices:
            return ""

        devices_str = "    devices:"
        for dev in devices.values():
            devices_str += f'\n      - "{dev}:{dev}"'
        return devices_str

    def _gen_ports(self) -> str:
        port_arg = getattr(self.args, "port", None)
        if not port_arg:
            # Default to 8080 if no port is specified
            return '    ports:\n      - "8080:8080"'

        p = port_arg.split(":", 2)
        host_port = p[1] if len(p) > 1 else p[0]
        container_port = p[0]
        return f'    ports:\n      - "{host_port}:{container_port}"'

    def _gen_environment(self) -> str:
        env_vars = get_accel_env_vars()
        # Allow user to override with --env
        if getattr(self.args, "env", None):
            for e in self.args.env:
                key, val = e.split("=", 1)
                env_vars[key] = val

        if not env_vars:
            return ""

        env_spec = "    environment:"
        for k, v in env_vars.items():
            env_spec += f'\n      - {k}={v}'
        return env_spec

    def _gen_gpu_deployment(self) -> str:
        gpu_keywords = ["cuda", "rocm", "gpu"]
        if not any(keyword in self.image.lower() for keyword in gpu_keywords):
            return ""

        return """\
    deploy:
      resources:
        reservations:
          devices:
            - driver: nvidia
              count: all
              capabilities: [gpu]"""

    def _gen_command(self) -> str:
        if not self.exec_args:
            return ""
        # shlex.join is perfect for creating a command string from a list
        cmd = shlex.join(self.exec_args)
        return f"    command: {cmd}"

    def generate(self) -> PlainFile:
        _version = version()

        # Generate all the dynamic sections of the YAML file
        volumes_string = self._gen_volumes()
        ports_string = self._gen_ports()
        environment_string = self._gen_environment()
        devices_string = self._gen_devices()
        gpu_deploy_string = self._gen_gpu_deployment()
        command_string = self._gen_command()

        # Assemble the final file content
        content = f"""\
# Save this output to a 'docker-compose.yaml' file and run 'docker compose up'.
#
# Created with ramalama-{_version}

services:
  {self.model_name}:
    container_name: {self.name}
    image: {self.image}
{volumes_string}
{ports_string}
{environment_string}
{devices_string}
{gpu_deploy_string}
{command_string}
    restart: unless-stopped
"""
        # Clean up any empty lines that might result from empty sections
        content = "\n".join(line for line in content.splitlines() if line.strip())

        return genfile(self.name, content)


def genfile(name: str, content: str) -> PlainFile:
    file_name = "docker-compose.yaml"
    print(f"Generating Compose YAML file: {file_name}")

    file = PlainFile(file_name)
    file.content = content
    return file
